from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import ContextManager, Union

from pharmpy.model import Model

from ...workflows.model_entry import ModelEntry
from ..results import Results


class ModelTransaction(ABC):
    @abstractmethod
    def store_model(self) -> None:
        """Store the model object bound to this transaction"""
        pass

    @abstractmethod
    def store_local_file(self, path: Path, new_filename: Union[str, None] = None):
        """Store a file from the local machine for the model bound to this
        transaction

        Parameters
        ----------
        path : Path
            Path to file
        new_filename: str|None
            Filename to give to the file. Optional, defaults to original
            filename given by path.
        """
        pass

    @abstractmethod
    def store_metadata(self, metadata) -> None:
        """Store metadata for the model bound to this transaction

        Parameters
        ----------
        metadata : Dict
            A dictionary with metadata
        """
        pass

    @abstractmethod
    def store_modelfit_results(self) -> None:
        """Store modelfit results of the model bound to this transaction"""
        pass

    @abstractmethod
    def store_model_entry(self) -> None:
        """Store model entry of the model entry bound to this transaction"""
        pass


class ModelSnapshot(ABC):
    @abstractmethod
    def retrieve_local_files(self, destination_path: Path) -> None:
        """Retrieve all files related to the model bound to this snapshot

        Parameters
        ----------
        destination_path : Path
            Local destination path
        """
        pass

    @abstractmethod
    def retrieve_file(self, filename: str) -> Path:
        """Retrieve one file related to the model bound to this snapshot

        Note that if the database is implemented in the local filesystem
        it is allowed to return a path directly to the file in the database.

        Parameters
        ----------
        filename : str
            Name of file

        Returns
        -------
        Path
            Path to local file
        """
        pass

    @abstractmethod
    def retrieve_model(self) -> Model:
        """Get the model bound to this snapshot

        Returns
        -------
        Model
            Retrieved model object
        """
        pass

    @abstractmethod
    def retrieve_modelfit_results(self) -> Results:
        """Read modelfit results from the database

        Parameters
        ----------
        name : str
            Name of the model

        Returns
        -------
        Results
            Retrieved model results object
        """
        pass

    @abstractmethod
    def retrieve_model_entry(self) -> ModelEntry:
        """Read model entry from the database

        Returns
        -------
        ModelEntry
            Retrieved model entry object
        """
        pass


class ModelDatabase(ABC):
    """Baseclass for databases for models and results of model runs

    Primary key is the model name, which means that models with
    the same name will be counted as the same model.

    Currently the model database is centered around storing and retrieving
    files belonging to the model or results from fitting the model. This
    doesn't mean that implementations have to use a file system to store the
    files.
    """

    @abstractmethod
    def store_model(self, model: Model):
        """Store a model object

        Parameters
        ----------
        model : Model
            Pharmpy model object
        """
        pass

    @abstractmethod
    def store_local_file(self, model: Model, path: Path, new_filename: Union[str, None] = None):
        """Store a file from the local machine

        Parameters
        ----------
        model : Model
            Pharmpy model object
        path : Path
            Path to file
        new_filename: str|None
            Filename to give to the file. Optional, defaults to original
            filename given by path.
        """
        pass

    @abstractmethod
    def store_metadata(self, model: Model, metadata: dict):
        """Store metadata

        Parameters
        ----------
        model : Model
            Pharmpy model object
        metadata : Dict
            A dictionary with metadata
        """
        pass

    @abstractmethod
    def store_modelfit_results(self, model: Model):
        """Store modelfit results

        Parameters
        ----------
        model : Model
            Pharmpy model object
        """
        pass

    @abstractmethod
    def store_model_entry(self, model_entry: ModelEntry) -> None:
        """Store model entry of the model entry bound to this transaction

        Parameters
        ----------
        model_entry : ModelEntry
            Pharmpy ModelEntry object
        """
        pass

    @abstractmethod
    def retrieve_local_files(self, model_name: str, destination_path: Path):
        """Retrieve all files related to a model

        Parameters
        ----------
        model_name : str
            Name of the model
        destination_path : Path
            Local destination path
        """
        pass

    @abstractmethod
    def retrieve_file(self, model_name: str, filename: str) -> Path:
        """Retrieve one file related to a model

        Note that if the database is implemented in the local filesystem
        it is allowed to return a path directly to the file in the database.

        Parameters
        ----------
        model_name : str
            Name of the model
        filename : str
            Name of file

        Returns
        -------
        Path
            Path to local file
        """
        pass

    @abstractmethod
    def retrieve_model(self, model_name: str) -> Model:
        """Read a model from the database

        Parameters
        ----------
        model_name : str
            Name of the model

        Returns
        -------
        Model
            Retrieved model object
        """
        pass

    @abstractmethod
    def retrieve_modelfit_results(self, model_name: str) -> Results:
        """Read modelfit results from the database

        Parameters
        ----------
        model_name : str
            Name of the model

        Returns
        -------
        Results
            Retrieved model results object
        """
        pass

    @abstractmethod
    def retrieve_model_entry(self, model_name: str) -> ModelEntry:
        """Read model entry from the database

        Parameters
        ----------
        model_name : str
            Name of the model

        Returns
        -------
        ModelEntry
            Retrieved model entry object
        """
        pass

    @abstractmethod
    def snapshot(self, model_name: str) -> ContextManager[ModelSnapshot]:
        """Creates a readable snapshot context for a given model.

        Parameters
        ----------
        model_name : str
            Name of the Pharmpy model object
        """
        pass

    @abstractmethod
    def transaction(
        self, model_or_model_entry: Union[Model, ModelEntry]
    ) -> ContextManager[ModelTransaction]:
        """Creates a writable transaction context for a given model.

        Parameters
        ----------
        model_or_model_entry : Model | ModelEntry
            Pharmpy model or ModelEntry object
        """
        pass


class NonTransactionalModelDatabase(ModelDatabase):
    @contextmanager
    def snapshot(self, model_name: str):
        yield DummySnapshot(self, model_name)

    @contextmanager
    def transaction(self, model_or_model_entry: Union[Model, ModelEntry]):
        yield DummyTransaction(self, model_or_model_entry)


class DummyTransaction(ModelTransaction):
    def __init__(self, database: ModelDatabase, model_or_model_entry: Union[Model, ModelEntry]):
        self.db = database
        if isinstance(model_or_model_entry, ModelEntry):
            self.model = model_or_model_entry.model
            self.model_entry = model_or_model_entry
        else:
            self.model = model_or_model_entry
            self.model_entry = None

    def store_model(self) -> None:
        return self.db.store_model(self.model)

    def store_local_file(self, path: Path, new_filename: Union[str, None] = None) -> None:
        return self.db.store_local_file(self.model, path, new_filename)

    def store_metadata(self, metadata: dict) -> None:
        return self.db.store_metadata(self.model, metadata)

    def store_modelfit_results(self) -> None:
        return self.db.store_modelfit_results(self.model)

    def store_model_entry(self) -> None:
        return self.db.store_model_entry(self.model_entry)


class DummySnapshot(ModelSnapshot):
    def __init__(self, database: ModelDatabase, model_name: str):
        self.db = database
        self.name = model_name

    def retrieve_local_files(self, destination_path: Path) -> None:
        return self.db.retrieve_local_files(self.name, destination_path)

    def retrieve_file(self, filename: str) -> Path:
        return self.db.retrieve_file(self.name, filename)

    def retrieve_model(self) -> Model:
        return self.db.retrieve_model(self.name)

    def retrieve_modelfit_results(self) -> Results:
        return self.db.retrieve_modelfit_results(self.name)

    def retrieve_model_entry(self) -> ModelEntry:
        return self.db.retrieve_model_entry(self.name)


class TransactionalModelDatabase(ModelDatabase):
    def store_model(self, model: Model) -> None:
        with self.transaction(model) as txn:
            return txn.store_model()

    def store_local_file(
        self, model: Model, path: Path, new_filename: Union[str, None] = None
    ) -> None:
        with self.transaction(model) as txn:
            return txn.store_local_file(path, new_filename)

    def store_metadata(self, model: Model, metadata: dict) -> None:
        with self.transaction(model) as txn:
            return txn.store_metadata(metadata)

    def store_modelfit_results(self, model: Model) -> None:
        with self.transaction(model) as txn:
            return txn.store_modelfit_results()

    def store_model_entry(self, model_entry: ModelEntry) -> None:
        with self.transaction(model_entry) as txn:
            return txn.store_model_entry()

    def retrieve_file(self, model_name: str, filename: str) -> Path:
        with self.snapshot(model_name) as sn:
            return sn.retrieve_file(filename)

    def retrieve_local_files(self, model_name: str, destination_path: Path) -> None:
        with self.snapshot(model_name) as sn:
            return sn.retrieve_local_files(destination_path)

    def retrieve_model(self, model_name: str) -> Model:
        with self.snapshot(model_name) as sn:
            return sn.retrieve_model()

    def retrieve_modelfit_results(self, model_name: str) -> Results:
        with self.snapshot(model_name) as sn:
            return sn.retrieve_modelfit_results()

    def retrieve_model_entry(self, model_name: str) -> ModelEntry:
        with self.snapshot(model_name) as sn:
            return sn.retrieve_model_entry()


class PendingTransactionError(Exception):
    pass



"""DataInfo is a companion to the dataset. It contains metadata of the dataset
"""
from __future__ import annotations

import json
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Tuple, Union, cast, overload

from pharmpy.internals.expr.units import parse as parse_units
from pharmpy.internals.fs.path import path_absolute, path_relative_to
from pharmpy.internals.immutable import Immutable, frozenmapping

if TYPE_CHECKING:
    import pandas as pd
    import sympy
else:
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy


class ColumnInfo(Immutable):
    """Information about one data column

    Parameters
    ----------
    name : str
        Colum name
    type : str
        Type (see the "type" attribute)
    unit : str
        Unit (see the "unit" attribute)
    scale : str
        Scale of measurement (see the "scale" attribute)
    continuous : bool
        True if continuous or False if discrete
    categories : Optional[Union[Tuple, Dict]]
        Tuple of all possible categories or dict from value to label for each category
    drop : bool
        Should column be dropped (i.e. barred from being used)
    datatype : str
        Pandas datatype or special Pharmpy datatype (see the "dtype" attribute)
    descriptor : str
        Descriptor (kind) of data
    """

    _all_types = (
        'id',
        'dv',
        'dvid',
        'idv',
        'unknown',
        'dose',
        'rate',
        'additional',
        'ii',
        'ss',
        'event',
        'covariate',
        'mdv',
        'compartment',
        'admid',
        'lloq',
        'blq',
    )
    _all_scales = ('nominal', 'ordinal', 'interval', 'ratio')
    _all_dtypes = (
        'int8',
        'int16',
        'int32',
        'int64',
        'uint8',
        'uint16',
        'uint32',
        'uint64',
        'float16',
        'float32',
        'float64',
        'float128',
        'nmtran-time',
        'nmtran-date',
        'str',
    )
    _all_descriptors = (
        None,
        'age',
        'body weight',
        'lean body mass',
        'fat free mass',
        'time after dose',
        'plasma concentration',
        'subject identifier',
        'observation identifier',
        'pk measurement',
        'pd measurement',
    )

    @staticmethod
    def convert_pd_dtype_to_datatype(dtype) -> str:
        """Convert pandas dtype to Pharmpy datatype

        Parameters
        ----------
        dtype : str
            String representing a pandas dtype

        Returns
        -------
        str
            String representing a Pharmpy datatype

        Examples
        --------
        >>> from pharmpy.model import ColumnInfo
        >>> ColumnInfo.convert_pd_dtype_to_datatype("float64")
        'float64'
        """
        return dtype if dtype in ColumnInfo._all_dtypes else 'str'

    @staticmethod
    def convert_datatype_to_pd_dtype(datatype) -> str:
        """Convert Pharmpy datatype to pandas dtype

        Parameters
        ----------
        datatype : str
            String representing a Pharmpy datatype

        Returns
        -------
        str
            String representing a pandas dtype

        Examples
        --------
        >>> from pharmpy.model import ColumnInfo
        >>> ColumnInfo.convert_datatype_to_pd_dtype("float64")
        'float64'
        >>> ColumnInfo.convert_datatype_to_pd_dtype("nmtran-date")
        'str'
        """
        if datatype.startswith('nmtran'):
            return 'str'
        else:
            return datatype

    def __init__(
        self,
        name: str,
        type: str = 'unknown',
        unit: sympy.Expr = sympy.Integer(1),
        scale: str = 'ratio',
        continuous: Optional[bool] = None,
        categories: Optional[Union[frozenmapping[str, str], tuple[str, ...]]] = None,
        drop: bool = False,
        datatype: str = "float64",
        descriptor: Optional[str] = None,
    ):
        self._name = name
        self._type = type
        self._unit = unit
        self._scale = scale
        self._continuous = continuous
        self._categories = categories
        self._drop = drop
        self._datatype = datatype
        self._descriptor = descriptor

    @staticmethod
    def _canonicalize_categories(
        categories: Union[Mapping[str, str], Sequence[str], None]
    ) -> Union[frozenmapping[str, str], tuple[str, ...], None]:
        if isinstance(categories, dict):
            return frozenmapping(categories)
        elif isinstance(categories, frozenmapping):
            return categories
        elif isinstance(categories, tuple):
            return categories
        elif isinstance(categories, Sequence):
            return tuple(categories)
        elif categories is None:
            return categories
        else:
            raise TypeError("categories must be None, list-like or dict-like")

    @classmethod
    def create(
        cls,
        name: str,
        type: str = 'unknown',
        unit: Optional[Union[str, int, float, sympy.Expr]] = None,
        scale: str = 'ratio',
        continuous: Optional[bool] = None,
        categories: Optional[Union[Mapping[str, str], Sequence[str]]] = None,
        drop: bool = False,
        datatype: str = "float64",
        descriptor: Optional[str] = None,
    ):
        if scale in ('nominal', 'ordinal'):
            if continuous is True:
                raise ValueError("A nominal or ordinal column cannot be continuous")
            else:
                continuous = False
        if continuous is None:
            continuous = True
        if not isinstance(name, str):
            raise TypeError("Column name must be a string")
        if type not in ColumnInfo._all_types:
            raise TypeError(f"Unknown column type {type}")
        if scale not in ColumnInfo._all_scales:
            raise TypeError(
                f"Unknown scale of measurement {scale}. Only {ColumnInfo._all_scales} are possible."
            )
        if unit is None:
            unit = sympy.Integer(1)
        else:
            unit = parse_units(unit)
        if datatype not in ColumnInfo._all_dtypes:
            raise ValueError(
                f"{datatype} is not a valid datatype. Valid datatypes are {ColumnInfo._all_dtypes}"
            )
        if descriptor not in ColumnInfo._all_descriptors:
            raise TypeError(f"Unknown column descriptor {descriptor}")
        categories = ColumnInfo._canonicalize_categories(categories)

        return cls(
            name=name,
            type=type,
            unit=unit,
            scale=scale,
            continuous=continuous,
            categories=categories,
            drop=drop,
            datatype=datatype,
            descriptor=descriptor,
        )

    def replace(self, **kwargs) -> ColumnInfo:
        """Replace properties and create a new ColumnInfo"""
        d = {key[1:]: value for key, value in self.__dict__.items()}
        d.update(kwargs)
        new = ColumnInfo.create(**d)
        return new

    def __eq__(self, other: Any):
        return (
            isinstance(other, ColumnInfo)
            and self._name == other._name
            and self._type == other._type
            and self._unit == other._unit
            and self._scale == other._scale
            and self._continuous == other._continuous
            and self._categories == other._categories
            and self._drop == other._drop
            and self._datatype == other._datatype
        )

    def __hash__(self):
        return hash(
            (
                self._name,
                self._type,
                self._unit,
                self._scale,
                self._continuous,
                # FIXME: What are categories really?
                # self._categories,
                self._drop,
                self._datatype,
                self._descriptor,
            )
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'name': self._name,
            'type': self._type,
            'unit': sympy.srepr(self._unit),
            'scale': self._scale,
            'continuous': self._continuous,
            'categories': self._categories,
            'drop': self._drop,
            'datatype': self._datatype,
            'descriptor': self._descriptor,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> ColumnInfo:
        unit = parse_units(d['unit'])
        assert isinstance(unit, sympy.Expr)
        return cls(
            name=d['name'],
            type=d['type'],
            unit=unit,
            scale=d['scale'],
            continuous=d['continuous'],
            categories=d['categories'],
            drop=d['drop'],
            datatype=d['datatype'],
            descriptor=d['descriptor'],
        )

    @property
    def name(self) -> str:
        """Column name"""
        return self._name

    @property
    def type(self) -> str:
        """Type of column

        ============  =============
        type          Description
        ============  =============
        id            Individual identifier. Max one per DataFrame. All values have to be unique
        idv           Independent variable. Max one per DataFrame.
        dv            Observations of the dependent variable
        dvid          Dependent variable ID
        covariate     Covariate
        dose          Dose amount
        rate          Rate of infusion
        additional    Number of additional doses
        ii            Interdose interval
        ss            Steady state dosing
        event         0 = observation
        mdv           0 = DV is observation value, 1 = DV is missing
        admid         Administration ID
        compartment   Compartment information (not yet exactly specified)
        lloq          Lower limit of quantification
        blq           Below limit of quantification indicator
        unknown       Unkown type. This will be the default for columns that hasn't been
                      assigned a type
        ============  =============
        """
        return self._type

    @property
    def descriptor(self) -> Union[str, None]:
        """Kind of data

        ====================== ============================================
        descriptor             Description
        ====================== ============================================
        age                    Age (since birth)
        body weight            Human body weight
        lean body mass         Lean body mass
        fat free mass          Fat free mass
        time after dose        Time after dose
        plasma concentration   Concentration of substance in blood plasma
        subject identifier     Unique integer identifier for a subject
        observation identifier Unique integer identifier for an observation
        pk measurement         Any kind of PK measurement
        pd measurement         Any kind of PD measurement
        ====================== ============================================
        """
        return self._descriptor

    @property
    def unit(self) -> sympy.Expr:
        """Unit of the column data

        Custom units are allowed, but units that are available in sympy.physics.units can be
        recognized. The default unit is 1, i.e. without unit.
        """
        return self._unit

    @property
    def scale(self) -> str:
        """Scale of measurement

        The statistical scale of measurement for the column data. Can be one of
        'nominal', 'ordinal', 'interval' and 'rational'.
        """
        return self._scale

    @property
    def continuous(self) -> Union[bool, None]:
        """Is the column data continuous

        True for continuous data and False for discrete. Note that nominal and ordinal data have to
        be discrete.
        """
        return self._continuous

    @property
    def categories(self) -> Union[frozenmapping[str, str], tuple[str, ...], None]:
        """List or dict of allowed categories"""
        return self._categories

    @property
    def drop(self) -> bool:
        """Should this column be dropped"""
        return self._drop

    @property
    def datatype(self) -> str:
        """Column datatype

        ============ ================ ======== ================================= ===========
        datatype     Description      Size     Range                             NA allowed?
        ============ ================ ======== ================================= ===========
        int8         Signed integer   8 bits   -128 to +127.                     No
        int16        Signed integer   16 bits  -32,768 to +32,767.               No
        int32        Signed integer   32 bits  -2,147,483,648 to +2,147,483,647. No
        int64        Signed integer   64 bits  -9,223,372,036,854,775,808 to     No
                                               9,223,372,036,854,775,807.
        uint8        Unsigned integer 8 bits   0 to 256.                         No
        uint16       Unsigned integer 16 bit   0 to 65,535.                      No
        uint32       Unsigned integer 32 bit   0 to 4,294,967,295.               No
        uint64       Unsigned integer 64 bit   0 to 18,446,744,073,709,551,615   No
        float16      Binary float     16 bits  ≈ ±6.55×10⁴                       Yes
        float32      Binary float     32 bits  ≈ ±3.4×10³⁸                       Yes
        float64      Binary float     64 bits  ≈ ±1.8×10³⁰⁸                      Yes
        float128     Binary float     128 bits ≈ ±1.2×10⁴⁹³²                     Yes
        nmtran-time  NM-TRAN time     n                                          No
        nmtran-date  NM-TRAN date     n                                          No
        str          General string   n                                          No
        ============ ================ ========================================== ===========

        The default, and most common datatype, is float64.
        """
        return self._datatype

    def is_categorical(self) -> bool:
        """Check if the column data is categorical

        Returns
        -------
        bool
            True if categorical (nominal or ordinal) and False otherwise.

        See also
        --------
        is_numerical : Check if the column data is numerical

        Examples
        --------
        >>> from pharmpy.model import ColumnInfo
        >>> col1 = ColumnInfo.create("WGT", scale='ratio')
        >>> col1.is_categorical()
        False
        >>> col2 = ColumnInfo.create("ID", scale='nominal')
        >>> col2.is_categorical()
        True

        """
        return self.scale in ['nominal', 'ordinal']

    def is_numerical(self) -> bool:
        """Check if the column data is numerical

        Returns
        -------
        bool
            True if numerical (interval or ratio) and False otherwise.

        See also
        --------
        is_categorical : Check if the column data is categorical

        Examples
        --------
        >>> from pharmpy.model import ColumnInfo
        >>> col1 = ColumnInfo.create("WGT", scale='ratio')
        >>> col1.is_numerical()
        True
        >>> col2 = ColumnInfo.create("ID", scale='nominal')
        >>> col2.is_numerical()
        False

        """
        return self.scale in ['interval', 'ratio']

    def is_integer(self) -> bool:
        """Check if the column datatype is integral

        Returns
        -------
        bool
            True if of integral datatype

        See also
        --------
        is_categorical : Check if the column data is categorical

        Examples
        --------
        >>> from pharmpy.model import ColumnInfo
        >>> col1 = ColumnInfo.create("WGT", scale='ratio')
        >>> col1.is_integer()
        False
        """
        return self.datatype in [
            'int8',
            'int16',
            'int32',
            'int64',
            'uint8',
            'uint16',
            'uint32',
            'uint64',
        ]

    def get_all_categories(self) -> list[str]:
        """Get a list of all categories"""
        if isinstance(self._categories, tuple):
            return list(self._categories)
        elif self._categories is None:
            return []
        else:
            return list(self._categories.keys())

    def __repr__(self):
        ser = pd.Series(
            [
                self._type,
                self._scale,
                self._continuous,
                self._categories,
                self._unit,
                self._drop,
                self._datatype,
                self._descriptor,
            ],
            index=[
                'type',
                'scale',
                'continuous',
                'categories',
                'unit',
                'drop',
                'datatype',
                'descriptor',
            ],
            name=self._name,
        )
        return ser.to_string(name=True)


class DataInfo(Sequence, Immutable):
    """Metadata for the dataset

    Can be indexed to get ColumnInfo for the columns.

    Parameters
    ----------
    columns : list
        List of column names
    path : Path
        Path to dataset file
    separator : str
        Character or regexp separator for dataset
    """

    def __init__(
        self,
        columns: tuple[ColumnInfo, ...] = (),
        path: Optional[Path] = None,
        separator: str = ',',
        force_absolute_path: bool = True,
    ):
        self._columns = columns
        self._path = path
        self._separator = separator
        self._force_absolute_path = force_absolute_path

    @classmethod
    def create(
        cls,
        columns: Optional[Union[Sequence[ColumnInfo], Sequence[str]]] = None,
        path: Optional[Union[str, Path]] = None,
        separator: str = ',',
        force_absolute_path: bool = True,
    ):
        if columns and not all(
            isinstance(col, str) or isinstance(col, ColumnInfo) for col in columns
        ):
            raise ValueError(
                'Argument `columns` need to consist of either type `str` or `ColumnInfo`'
            )
        if columns is None or len(columns) == 0:
            cols: Tuple[ColumnInfo, ...] = ()
        elif len(columns) > 0 and any(isinstance(col, str) for col in columns):
            cols = tuple(ColumnInfo.create(col) if isinstance(col, str) else col for col in columns)
        else:
            cols = cast(Tuple[ColumnInfo, ...], tuple(columns))
        assert isinstance(cols, tuple)
        if path is not None:
            path = Path(path)
        assert not force_absolute_path or path is None or path.is_absolute()
        return cls(columns=cols, path=path, separator=separator)

    def replace(self, **kwargs) -> DataInfo:
        if 'columns' in kwargs:
            columns = tuple(kwargs['columns'])
        else:
            columns = self._columns
        if 'path' in kwargs:
            if kwargs['path'] is not None:
                path = Path(kwargs['path'])
            else:
                path = None
        else:
            path = self._path
        separator = kwargs.get('separator', self._separator)
        return DataInfo.create(columns=columns, path=path, separator=separator)

    def __add__(self, other: DataInfo) -> DataInfo:
        if isinstance(other, DataInfo):
            return DataInfo.create(
                columns=self._columns + other._columns, path=self.path, separator=self.separator
            )
        elif isinstance(other, ColumnInfo):
            return DataInfo.create(
                columns=self._columns + (other,), path=self.path, separator=self.separator
            )
        else:
            return DataInfo.create(
                columns=self._columns + tuple(other), path=self.path, separator=self.separator
            )

    def __radd__(self, other: DataInfo) -> DataInfo:
        if isinstance(other, ColumnInfo):
            return DataInfo.create(
                columns=(other,) + self._columns, path=self.path, separator=self.separator
            )
        else:
            return DataInfo.create(
                columns=tuple(other) + self._columns, path=self.path, separator=self.separator
            )

    def __eq__(self, other: Any):
        if not isinstance(other, DataInfo):
            return False
        if len(self) != len(other):
            return False
        for col1, col2 in zip(self, other):
            if col1 != col2:
                return False
        return True

    def __hash__(self):
        return hash(self._columns)

    def __len__(self):
        return len(self._columns)

    def _getindex(self, i: Union[int, str]) -> int:
        if isinstance(i, str):
            for n, col in enumerate(self._columns):
                if col.name == i:
                    return n
            raise IndexError(f"Cannot find column {i} in DataInfo")
        elif isinstance(i, int):
            return i
        else:
            raise TypeError(f"Cannot index DataInfo by {type(i)}")

    @overload
    def __getitem__(self, index: Union[list, slice]) -> DataInfo:
        ...

    @overload
    def __getitem__(self, index: Union[int, str]) -> ColumnInfo:
        ...

    def __getitem__(  # pyright: ignore [reportIncompatibleMethodOverride]
        self, index: Union[list, slice, int, str]
    ) -> Union[DataInfo, ColumnInfo]:
        if isinstance(index, list):
            cols = []
            for ind in index:
                i = self._getindex(ind)
                cols.append(self._columns[i])
            return DataInfo.create(columns=cols)
        if isinstance(index, slice):
            return DataInfo.create(self._columns[index], path=self._path, separator=self._separator)

        return self._columns[self._getindex(index)]

    @property
    def path(self) -> Union[Path, None]:
        r"""Path of dataset file

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> str(model.datainfo.path).replace('\\', '/')     # doctest: +ELLIPSIS
        '.../pharmpy/modeling/example_models/pheno.dta'
        """
        return self._path

    @property
    def separator(self) -> str:
        """Separator for dataset file

        Can be a single character or a regular expression
        string.
        """
        return self._separator

    @property
    def typeix(self) -> TypeIndexer:
        """Type indexer

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.typeix['covariate'].names
        ['WGT', 'APGR']
        """
        return TypeIndexer(self)

    @property
    def descriptorix(self) -> DescriptorIndexer:
        """Descriptor indexer

        Example
        -------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.descriptorix['body weight'].names
        ['WGT']
        """
        return DescriptorIndexer(self)

    def set_column(self, col: ColumnInfo) -> DataInfo:
        """Set ColumnInfo of an existing column of the same name

        Parameters
        ----------
        col : ColumnInfo
            New ColumnInfo

        Returns
        -------
        DataInfo
            Updated DataInfo
        """
        newcols = []
        for cur in self:
            if cur.name != col.name:
                newcols.append(cur)
            else:
                newcols.append(col)
        return self.replace(columns=newcols)

    @property
    def id_column(self) -> ColumnInfo:
        """The id column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.id_column.name
        'ID'
        """
        return self.typeix['id'][0]

    def _set_column_type(self, name: str, type: str) -> DataInfo:
        for i, col in enumerate(self):
            if col.name != name and col.type == type:
                raise ValueError(
                    f"Cannot set new {type} column: column {col.name} already has type {type}"
                )

        for i, col in enumerate(self):
            if col.name == name:
                mycol = col
                ind = i
                break
        else:
            raise IndexError(f"No column {name} in DataInfo")

        newcol = mycol.replace(type=type)
        cols = self._columns[0:ind] + (newcol,) + self._columns[ind + 1 :]
        return DataInfo.create(cols, path=self._path, separator=self._separator)

    def set_id_column(self, name: str) -> DataInfo:
        return self._set_column_type(name, 'id')

    @property
    def dv_column(self) -> ColumnInfo:
        """The dv column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.dv_column.name
        'DV'
        """
        return self.typeix['dv'][0]

    def set_dv_column(self, name: str) -> DataInfo:
        return self._set_column_type(name, 'dv')

    @property
    def idv_column(self) -> ColumnInfo:
        """The idv column

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.idv_column.name
        'TIME'
        """
        return self.typeix['idv'][0]

    def set_idv_column(self, name: str) -> DataInfo:
        return self._set_column_type(name, 'idv')

    @property
    def names(self) -> list[str]:
        """All column names

        Examples
        --------
        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.datainfo.names
        ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV', 'FA1', 'FA2']
        """
        return [col.name for col in self._columns]

    @property
    def types(self) -> list[str]:
        """All column types"""
        return [col.type for col in self._columns]

    def set_types(self, value: Union[list[str], str]) -> DataInfo:
        """Set types for all columns

        Parameters
        ----------
        value : list or str
            Types to set. If only one this will be broadcast

        Return
        ------
        DataInfo
            Updated datainfo
        """
        if isinstance(value, str):
            value = [value]
        if len(value) == 1:
            value *= len(self)
        if len(value) != len(self):
            raise ValueError(
                "Length mismatch. "
                "Can only set the same number of names as columns or 1 for broadcasting"
            )
        newcols = []
        for v, col in zip(value, self._columns):
            newcol = col.replace(type=v)
            newcols.append(newcol)
        return DataInfo.create(columns=newcols, path=self._path, separator=self._separator)

    def get_dtype_dict(self) -> dict[str, str]:
        """Create a dictionary from column names to pandas dtypes

        This can be used as input to some pandas functions to convert
        column to the correct pandas dtype.

        Returns
        -------
        dict
            Column name to pandas dtype

        Examples
        --------
        >>> from pharmpy.modeling import *
        >>> model = load_example_model("pheno")
        >>> model.datainfo.get_dtype_dict()
        {'ID': 'int32',
         'TIME': 'float64',
         'AMT': 'float64',
         'WGT': 'float64',
         'APGR': 'float64',
         'DV': 'float64',
         'FA1': 'float64',
         'FA2': 'float64'}
        """
        return {
            col.name: col.datatype
            if not col.drop and not col.datatype.startswith('nmtran')
            else 'str'
            for col in self
        }

    def to_dict(self) -> dict[str, Any]:
        return self._to_dict(path=None)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> DataInfo:
        columns = tuple(ColumnInfo.from_dict(col) for col in d['columns'])
        return cls(columns=columns, path=d['path'], separator=d['separator'])

    def _to_dict(self, path: Optional[str]) -> dict[str, Any]:
        a = []
        for col in self._columns:
            d = {
                "name": col.name,
                "type": col.type,
                "scale": col.scale,
                "continuous": col.continuous,
                "categories": col.categories,
                "unit": str(col.unit),
                "datatype": col.datatype,
                "drop": col.drop,
                "descriptor": col.descriptor,
            }
            a.append(d)

        return {
            "columns": a,
            "path": path,
            "separator": self._separator,
        }

    def to_json(self, path: Optional[Union[Path, str]] = None):
        if path is None:
            return json.dumps(self._to_dict(str(self.path) if self.path is not None else None))
        else:
            with open(path, 'w') as fp:
                json.dump(
                    self._to_dict(
                        str(path_relative_to(Path(path).parent, self.path))
                        if self.path is not None
                        else None
                    ),
                    fp,
                )

    @staticmethod
    def from_json(s: str) -> DataInfo:
        """Create DataInfo from JSON string

        Parameters
        ----------
        s : str
            JSON string

        Return
        ------
        DataInfo
            Created DataInfo object
        """
        d = json.loads(s)
        columns = []
        for col in d['columns']:
            ci = ColumnInfo.create(
                name=col['name'],
                type=col.get('type', 'unknown'),
                scale=col['scale'],
                continuous=col.get('continuous', None),
                unit=col.get('unit', sympy.Integer(1)),
                categories=col.get('categories', None),
                datatype=col.get('datatype', 'float64'),
                descriptor=col.get('descriptor', None),
                drop=col.get('drop', False),
            )
            columns.append(ci)
        path = d.get('path', None)
        if path:
            path = Path(path)
        separator = d.get('separator', ',')
        di = DataInfo.create(columns, path=path, separator=separator, force_absolute_path=False)
        return di

    @staticmethod
    def read_json(path: Union[Path, str]) -> DataInfo:
        """Read DataInfo from JSON file

        Parameters
        ----------
        path : Path or str
            Path to JSON datainfo file

        Return
        ------
        DataInfo
            Created DataInfo object
        """
        with open(path, 'r') as fp:
            s = fp.read()
        di = DataInfo.from_json(s)
        return (
            di
            if di.path is None or di.path.is_absolute()
            else di.replace(path=path_absolute(Path(path).parent / di.path))
        )

    def __repr__(self):
        labels = [col.name for col in self._columns]
        types = [col.type for col in self._columns]
        scales = [col.scale for col in self._columns]
        cont = [col.continuous for col in self._columns]
        cats = [col.categories for col in self._columns]
        units = [col.unit for col in self._columns]
        drop = [col.drop for col in self._columns]
        datatype = [col.datatype for col in self._columns]
        descriptor = [col.descriptor for col in self._columns]
        df = pd.DataFrame(
            {
                'name': labels,
                'type': types,
                'scale': scales,
                'continuous': cont,
                'categories': cats,
                'unit': units,
                'drop': drop,
                'datatype': datatype,
                'descriptor': descriptor,
            }
        )
        return df.to_string(index=False)


class TypeIndexer:
    def __init__(self, obj):
        self._obj = obj

    def __getitem__(self, i) -> DataInfo:
        cols = [col for col in self._obj if col.type == i and not col.drop]
        if not cols:
            raise IndexError(f"No columns of type {i} available")
        return DataInfo.create(cols)


class DescriptorIndexer:
    def __init__(self, obj):
        self._obj = obj

    def __getitem__(self, i) -> DataInfo:
        cols = [col for col in self._obj if col.descriptor == i and not col.drop]
        if not cols:
            raise IndexError(f"No columns with descriptor {i} available")
        return DataInfo.create(cols)



from __future__ import annotations

import importlib
import json
import lzma
import re
import warnings
from contextlib import closing
from dataclasses import dataclass
from io import StringIO
from lzma import open as lzma_open
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union, overload

import pharmpy
from pharmpy.deps import altair as alt
from pharmpy.internals.immutable import Immutable
from pharmpy.model import Model

if TYPE_CHECKING:
    import pandas as pd

    from pharmpy.workflows import Log
else:
    from pharmpy.deps import pandas as pd


def mfr(model: Model) -> ModelfitResults:
    res = model.modelfit_results
    assert isinstance(res, ModelfitResults)
    return res


def _df_to_json(df: pd.DataFrame) -> dict[str, Any]:
    if str(df.columns.dtype) == 'int64':
        # Workaround for https://github.com/pandas-dev/pandas/issues/46392
        df.columns = df.columns.map(str)
    # Set double precision to 15 to remove some round-trip errors, however 17 should be set when its possible
    # See: https://github.com/pandas-dev/pandas/issues/38437
    df_json = df.to_json(orient='table', double_precision=15)
    assert df_json is not None
    return json.loads(df_json)


def _index_to_json(index: Union[pd.Index, pd.MultiIndex]) -> dict[str, Any]:
    if isinstance(index, pd.MultiIndex):
        return {'__class__': 'MultiIndex', **_df_to_json(index.to_frame(index=False))}
    return {'__class__': 'Index', **_df_to_json(index.to_frame(index=False))}


class ResultsJSONEncoder(json.JSONEncoder):
    def default(self, obj) -> Union[dict[str, Any], None]:
        # NOTE: This function is called when the base JSONEncoder does not know
        # how to encode the given object, so it will not be called on int,
        # float, str, list, tuple, and dict. It could be called on set for
        # instance, or any custom class.
        from pharmpy.workflows import LocalDirectoryToolDatabase, Log

        if isinstance(obj, Results):
            d = obj.to_dict()
            d['__module__'] = obj.__class__.__module__
            d['__class__'] = obj.__class__.__qualname__
            return d
        elif isinstance(obj, pd.DataFrame):
            d = _df_to_json(obj)
            d['__class__'] = 'DataFrame'
            return d
        elif isinstance(obj, pd.Series):
            if obj.size >= 1 and isinstance(obj.iloc[0], pd.DataFrame):
                # NOTE: Hack special case for Series of DataFrame objects
                return {
                    'data': [{'__class__': 'DataFrame', **_df_to_json(df)} for df in obj.values],
                    'index': _index_to_json(obj.index),
                    'name': obj.name,
                    'dtype': str(obj.dtype),
                    '__class__': 'Series[DataFrame]',
                }

            # NOTE: Hack to work around poor support of to_json/read_json of
            # pd.Series with MultiIndex
            df = obj.to_frame()
            d = _df_to_json(df)
            d['__class__'] = 'Series'
            return d
        elif obj.__class__.__module__.startswith('altair.'):
            with warnings.catch_warnings():
                # FIXME: Remove filter once altair stops relying on deprecated APIs
                warnings.filterwarnings(
                    "ignore",
                    message=".*iteritems is deprecated and will be removed in a future version. Use .items instead.",
                    category=FutureWarning,
                )
                # This was fixed in altair v2.1.1 can be removed once we require that version
                warnings.filterwarnings(
                    "ignore",
                    message=".*the convert_dtype parameter is deprecated",
                    category=FutureWarning,
                )
                d = obj.to_dict()
            d['__module__'] = obj.__class__.__module__
            d['__class__'] = obj.__class__.__qualname__
            return d
        elif isinstance(obj, Model):
            # TODO: Consider using other representation, e.g. path
            return None
        elif isinstance(obj, Log):
            d: Dict[Any, Any] = obj.to_dict()
            d['__class__'] = obj.__class__.__qualname__
            return d
        elif isinstance(obj, LocalDirectoryToolDatabase):
            d = obj.to_dict()
            d['__module__'] = obj.__class__.__module__
            d['__class__'] = obj.__class__.__qualname__
            return d
        elif isinstance(obj, Path):
            d = {'path': str(obj), '__class__': 'PosixPath'}
            return d
        else:
            # NOTE: This will raise a proper TypeError
            return super().default(obj)


def _df_read_json(obj) -> pd.DataFrame:
    # Convert time strings to naive datetime and then to string
    # Needed because of https://github.com/pandas-dev/pandas/issues/52595
    for row in obj['data']:
        if 'time' in row:
            row['time'] = pd.to_datetime(row['time']).tz_localize(None)
            row['time'] = row['time'].isoformat()

    return pd.read_json(StringIO(json.dumps(obj)), typ='frame', orient='table', precise_float=True)


def _multi_index_read_json(obj) -> pd.MultiIndex:
    return pd.MultiIndex.from_frame(_df_read_json(obj))


class ResultsJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        # NOTE: This hook will be called for every dict produced by the
        # base JSONDecoder. It will not be called on int, float, str, or list.
        module = None
        cls = None

        if '__module__' in obj:
            module = obj['__module__']
            del obj['__module__']

        if '__class__' in obj:
            cls = obj['__class__']
            del obj['__class__']

        # NOTE: Handling cls not None and module is None is kept for backwards
        # compatibility

        if cls is None and module is not None:
            raise ValueError('Cannot specify module without specifying class')

        if module is None or module.startswith('pandas.'):
            if cls == 'DataFrame':
                return _df_read_json(obj)
            elif cls == 'Series':
                # NOTE: Hack to work around poor support of to_json/read_json of
                # pd.Series with MultiIndex
                df = _df_read_json(obj)
                series = df.iloc[:, 0]  # NOTE: First and only column.
                return series
            elif cls == 'Series[DataFrame]':
                # NOTE: Hack to work around poor support of Series of DataFrame
                # objects. All subobjects have already been converted.
                return pd.Series(
                    obj['data'], index=obj['index'], dtype=obj['dtype'], name=obj['name']
                )
            elif cls == 'Index':
                return _multi_index_read_json(obj).get_level_values(0)
            elif cls == 'MultiIndex':
                return _multi_index_read_json(obj)

        if module is None:
            if cls == 'vega-lite':
                # NOTE: Slow parsing for parsing PsN frem output and old format
                return alt.Chart.from_dict(obj, validate=True)

        if module is not None and module.startswith('altair.'):
            # NOTE: Fast parsing when reading own output
            assert cls is not None
            try:
                class_ = getattr(alt, cls)
            except AttributeError:
                raise ValueError(f'Unknown class {cls} in {module}')
            return class_.from_dict(obj, validate=False)

        if cls is not None and cls.endswith('Results'):
            if module is None:
                # NOTE: Kept for backwards compatibility: we guess the module
                # path based on the class name.
                tool_name = cls[:-7].lower()  # NOTE: Trim "Results" suffix
                tool_module = importlib.import_module(f'pharmpy.tools.{tool_name}')
                results_class = tool_module.results_class
            else:
                tool_module = importlib.import_module(module)
                results_class = getattr(tool_module, cls)

            return results_class.from_dict(obj)

        from pharmpy.workflows import LocalDirectoryToolDatabase, Log

        if cls is not None and cls == 'LocalDirectoryToolDatabase':
            return LocalDirectoryToolDatabase.from_dict(obj)

        if cls == 'PosixPath':
            return Path(obj)
        if cls == 'Log':
            return Log.from_dict(obj)

        return obj


def _is_likely_to_be_json(source: str):
    # NOTE: Heuristic to determine if path or buffer: first non-space character
    # is '{'.
    match = re.match(r'\s*([^\s])', source)
    return match is not None and match.group(1) == '{'


def read_results(path_or_str: Union[str, Path]):
    if isinstance(path_or_str, str) and _is_likely_to_be_json(path_or_str):
        manager = closing(StringIO(path_or_str))
    else:
        path = Path(path_or_str)
        if path.is_dir():
            path /= 'results.json'

        if path.name.endswith('.xz'):
            manager = lzma.open(path, 'r', encoding='utf-8')
        else:
            manager = open(path, 'r')

    with manager as readable:
        return json.load(readable, cls=ResultsJSONDecoder)


@dataclass(frozen=True)
class Results(Immutable):
    """Base class for all result classes"""

    __version__: str = pharmpy.__version__  # NOTE: Default version if not overridden

    @classmethod
    def from_dict(cls, d: dict[str, Any]):
        """Create results object from dictionary"""
        removed_keys = {
            '__version__',
            'best_model',  # NOTE: Was removed in d5b3503 and 8578c8b
            'input_model',  # NOTE: Was removed in d5b3503 and 8578c8b
        }
        return cls(
            __version__=d.get('__version__', 'unknown'),  # NOTE: Override default version
            **{k: v for k, v in d.items() if k not in removed_keys},
        )

    @overload
    def to_json(self, path: None = None, lzma: Literal[False] = False) -> str:
        ...

    @overload
    def to_json(self, path: Path, lzma: bool = False) -> None:
        ...

    def to_json(self, path: Optional[Path] = None, lzma: bool = False) -> Union[str, None]:
        """Serialize results object as json

        Parameters
        ----------
        path : Path
            Path to save json file or None to serialize to string
        lzma : bool
            Set to compress file with lzma

        Returns
        -------
        str
            Json as string unless path was used
        """
        s = ResultsJSONEncoder().encode(self)
        if path:
            if not lzma:
                with open(path, 'w') as fh:
                    fh.write(s)
            else:
                xz_path = path.parent / (path.name + '.xz')
                with lzma_open(xz_path, 'w') as fh:
                    fh.write(bytes(s, 'utf-8'))
        else:
            return s

    def get_and_reset_index(self, attr: str, **kwargs) -> pd.DataFrame:
        """Wrapper to reset index of attribute or result from method.

        Used to facilitate importing multiindex dataframes into R
        """
        val = getattr(self, attr)
        if callable(val):
            df = val(**kwargs)
        else:
            df = val
        return df.reset_index()

    def to_dict(self) -> dict[str, Any]:
        """Convert results object to a dictionary"""
        return vars(self).copy()

    def __str__(self):
        start = self.__class__.__name__
        s = f'{start}\n\n'
        d = self.to_dict()
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                s += value.to_string()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    df = pd.DataFrame(value)
                    df_str = df.to_string(index=False)
                    df_str = df_str.split('\n')[1:]
                    s += '\n'.join(df_str)
            else:
                s += str(value) + '\n'
            s += '\n\n'
        return s

    def to_csv(self, path: Path):
        """Save results as a human readable csv file

        Index will not be printed if it is a basic range.

        Parameters
        ----------
        path : Path
            Path to csv-file
        """
        d = self.to_dict()
        s = ""
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            elif isinstance(value, Model):
                continue
            elif isinstance(value, list) and isinstance(value[0], Model):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                if isinstance(value.index, pd.RangeIndex):
                    use_index = False
                else:
                    use_index = True
                csv = value.to_csv(index=use_index)
                assert isinstance(csv, str)
                s += csv
            elif isinstance(value, pd.Series):
                csv = value.to_csv()
                assert isinstance(csv, str)
                s += csv
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    for row in value:
                        s += f'{",".join(map(str, row))}\n'
            else:
                s += str(value) + '\n'
            s += '\n'
        with open(path, 'w', newline='') as fh:
            print(s, file=fh)


@dataclass(frozen=True)
class ModelfitResults(Results):
    """Base class for results from a modelfit operation

    Attributes
    ----------
    name : str
        Name of model
    description : str
        Description of model
    correlation_matrix : pd.DataFrame
        Correlation matrix of the population parameter estimates
    covariance_matrix : pd.DataFrame
        Covariance matrix of the population parameter estimates
    precision_matrix : pd.DataFrame
        Precision matrix of the population parameter estimates
    evaluation_ofv : float
        The objective function value as if the model was evaluated. Currently
        works for classical estimation methods by taking the OFV of the first
        iteration.
    individual_ofv : pd.Series
        OFV for each individual
    individual_estimates : pd.DataFrame
        Estimates for etas
    individual_estimates_covariance : pd.Series
        Estimated covariance between etas
    parameter_estimates : pd.Series
        Population parameter estimates
    parameter_estimates_iterations : pd.DataFrame
        All recorded iterations for parameter estimates
    parameter_estimates_sdcorr : pd.Series
        Population parameter estimates with variability parameters as standard deviations and
        correlations
    residuals: pd.DataFrame
        Table of various residuals
    predictions: pd.DataFrame
        Table of various predictions
    estimation_runtime : float
        Runtime for one estimation step
    runtime_total : float
        Total runtime of estimation
    standard_errors : pd.Series
        Standard errors of the population parameter estimates
    standard_errors_sdcorr : pd.Series
        Standard errors of the population parameter estimates on standard deviation and correlation
        scale
    relative_standard_errors : pd.Series
        Relative standard errors of the population parameter estimates
    termination_cause : str
        The cause of premature termination. One of 'maxevals_exceeded' and 'rounding_errors'
    function_evaluations : int
        Number of function evaluations
    evaluation : pd.Series
        A bool for each estimation step. True if this was a model evaluation and False otherwise
    covstep_successful : bool or None
        Covariance status.
    gradients : pd.Series
        Final parameter gradients
    gradients_iterations : pd.DataFrame
        All recorded parameter gradients
    warnings : List
        List of warnings
    """

    name: Optional[str] = None
    description: Optional[str] = None
    ofv: Optional[float] = None
    ofv_iterations: Optional[pd.Series] = None
    parameter_estimates: Optional[pd.Series] = None
    parameter_estimates_sdcorr: Optional[pd.Series] = None
    parameter_estimates_iterations: Optional[pd.DataFrame] = None
    covariance_matrix: Optional[pd.DataFrame] = None
    correlation_matrix: Optional[pd.DataFrame] = None
    precision_matrix: Optional[pd.DataFrame] = None
    standard_errors: Optional[pd.Series] = None
    standard_errors_sdcorr: Optional[pd.Series] = None
    relative_standard_errors: Optional[pd.Series] = None
    minimization_successful: Optional[bool] = None
    minimization_successful_iterations: Optional[pd.DataFrame] = None
    estimation_runtime: Optional[float] = None
    estimation_runtime_iterations: Optional[pd.DataFrame] = None
    individual_ofv: Optional[pd.Series] = None
    individual_estimates: Optional[pd.DataFrame] = None
    individual_estimates_covariance: Optional[pd.DataFrame] = None
    residuals: Optional[pd.DataFrame] = None
    predictions: Optional[pd.DataFrame] = None
    runtime_total: Optional[float] = None
    termination_cause: Optional[str] = None
    termination_cause_iterations: Optional[pd.Series] = None
    function_evaluations: Optional[float] = None
    function_evaluations_iterations: Optional[pd.Series] = None
    significant_digits: Optional[float] = None
    significant_digits_iterations: Optional[pd.Series] = None
    log_likelihood: Optional[float] = None
    log: Optional['Log'] = None
    evaluation: Optional[pd.Series] = None
    covstep_successful: Optional[bool, None] = None
    gradients: Optional[pd.Series] = None
    gradients_iterations: Optional[pd.DataFrame] = (None,)
    warnings: Optional[List[str]] = None

    def __repr__(self):
        return f'<Pharmpy modelfit results object {self.name}>'



from __future__ import annotations

from functools import reduce
from operator import __and__
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def assume_all(predicate: sympy.assumptions.Predicate, expressions: Iterable[sympy.Expr]):
    tautology = sympy.Q.is_true(True)
    return reduce(__and__, map(predicate, expressions), tautology)



import os
import re
import shutil
import ssl
import warnings
from pathlib import Path
from urllib.request import urlopen

from bs4 import BeautifulSoup
from csscompressor import compress
from sphinx.application import Sphinx

from pharmpy.internals.fs.cwd import chdir
from pharmpy.internals.fs.tmp import TemporaryDirectory


def generate_report(rst_path, results_path, target_path):
    """Generate report from rst and results json"""
    results_path = Path(results_path)
    with TemporaryDirectory() as tmpdirname:
        tmp_path = Path(tmpdirname)
        source_path = tmp_path / 'source'
        source_path.mkdir()
        shutil.copy(rst_path, source_path / 'results.rst')
        if results_path.is_dir():
            results_path /= 'results.json'
        shutil.copy(results_path, source_path)

        conf_path = Path(__file__).resolve().parent

        # Change directory for results.json to be found
        with chdir(source_path):
            with open(os.devnull, 'w') as devnull:
                with warnings.catch_warnings():
                    # Don't display deprecation warnings.
                    # See https://github.com/pharmpy/pharmpy/issues/20
                    warnings.filterwarnings("ignore", message="The app.add_stylesheet")
                    warnings.filterwarnings("ignore", message="The app.add_javascript")
                    # Deprecation warning in jupyter_sphinx
                    warnings.filterwarnings("ignore", message="Passing a schema to Validator")
                    # Deprecation warning in python 3.10
                    warnings.filterwarnings(
                        "ignore",
                        message="The distutils package is deprecated and slated for removal in "
                        "Python 3.12. Use setuptools or check PEP 632 for potential alternatives",
                    )
                    warnings.filterwarnings("ignore", "There is no current event loop")
                    # From jupyter-core 5.1.2
                    warnings.filterwarnings(
                        "ignore", "Jupyter is migrating its paths to use standard platformdirs"
                    )
                    warnings.filterwarnings(
                        "ignore", "The alias 'sphinx.util.progress_message' is deprecated"
                    )
                    warnings.filterwarnings(
                        "ignore", "nodes.Node.traverse\\(\\) is obsoleted by Node.findall\\(\\)."
                    )
                    # From Python 3.11
                    warnings.filterwarnings(
                        "ignore", "'imghdr' is deprecated and slated for removal in Python 3.13"
                    )
                    warnings.filterwarnings(
                        "ignore",
                        "zmq.eventloop.ioloop is deprecated in pyzmq 17.",
                    )

                    app = Sphinx(
                        str(source_path),
                        str(conf_path),
                        str(tmp_path),
                        str(tmp_path),
                        "singlehtml",
                        status=devnull,
                        warning=devnull,
                    )
                    app.build()

        # Write missing altair css
        with open(tmp_path / '_static' / 'altair-plot.css', 'w') as dh:
            dh.write(
                """.vega-actions a {
    margin-right: 12px;
    color: #757575;
    font-weight: normal;
    font-size: 13px;
}

.vega-embed {
    margin-bottom: 20px;
    margin-top: 20px;
    width: 100%;
}
"""
            )
        report_path = tmp_path / 'results.html'
        embed_css_and_js(tmp_path / 'results.html', report_path)
        shutil.copy(report_path, target_path)


def embed_css_and_js(html, target):
    """Embed all external css and javascript into an html"""
    with open(html, 'r', encoding='utf-8') as sh:
        soup = BeautifulSoup(sh, features='lxml')

    scripts = soup.findAll("script", attrs={"src": True})

    for script in scripts:
        source = script.attrs['src']
        if source.startswith('http'):
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE  # To avoid CERTIFICATE_VERIFY_FAILED
            with urlopen(source, context=ctx) as infile:
                content = infile.read().decode('utf-8')
        else:
            path = html.parent / source
            if path.name == 'thebelab-helper.js':  # This file wasn't created
                continue
            with open(path, 'r') as sh:
                content = sh.read()

        # Minification with jsmin didn't work
        tag = soup.new_tag('script')
        tag['type'] = 'text/javascript'
        tag.append(content)
        script.replace_with(tag)

    stylesheets = soup.findAll("link", attrs={"rel": "stylesheet"})

    for stylesheet in stylesheets:
        stylesheet_src = stylesheet.attrs['href']
        tag = soup.new_tag("style")
        tag['type'] = 'text/css'
        path = html.parent / stylesheet_src
        if path.name == 'thebelab.css':  # This file wasn't created
            continue
        with open(path, 'r') as sh:
            content = sh.read()
        if '@import' in content:
            import_files = re.findall(r'@import\s+url\("([A-Za-z0-9.]+)"\)', content)
            for name in import_files:
                with open(path.parent / name, 'r') as import_file:
                    import_content = import_file.read()
                content = re.sub(r'@import\s+url\("' + name + r'"\);', import_content, content)
        minified_content = compress(content)
        tag.append(minified_content)
        stylesheet.replace_with(tag)

    with open(target, 'w', encoding='utf-8') as dh:
        dh.write(str(soup))



from __future__ import annotations

import re
from dataclasses import dataclass
from itertools import filterfalse
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from pharmpy.internals.expr.assumptions import assume_all
from pharmpy.internals.expr.leaves import free_images_and_symbols
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.graph.directed.connected_components import strongly_connected_component_of
from pharmpy.internals.graph.directed.inverse import inverse as graph_inverse
from pharmpy.internals.graph.directed.reachability import reachable_from
from pharmpy.model import (
    Assignment,
    Compartment,
    CompartmentalSystem,
    Model,
    Statement,
    Statements,
    output,
)

from .parameters import get_omegas, get_sigmas, get_thetas

if TYPE_CHECKING:
    import networkx as nx
    import sympy
else:
    from pharmpy.deps import networkx as nx
    from pharmpy.deps import sympy

T = TypeVar('T')
U = TypeVar('U')


def get_observation_expression(model: Model):
    """Get the full symbolic expression for the observation according to the model

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_observation_expression
    >>> import sympy
    >>> model = load_example_model("pheno_linear")
    >>> expr = get_observation_expression(model)
    >>> sympy.pprint(expr)
    D_EPSETA1_2⋅EPS₁⋅(ETA₂ - OETA₂) + D_ETA1⋅(ETA₁ - OETA₁) + D_ETA2⋅(ETA₂ - OETA₂
    ) + EPS₁⋅(D_EPS1 + D_EPSETA1_1⋅(ETA₁ - OETA₁)) + OPRED
    """
    stats = model.statements
    # FIXME: Handle other DVs
    dv = list(model.dependent_variables.keys())[0]
    for i, s in enumerate(stats):
        if s.symbol == dv:
            y = s.expression
            break
    else:
        raise ValueError('Could not locate dependent variable expression')

    for j in range(i, -1, -1):
        y = subs(y, {stats[j].symbol: stats[j].expression}, simultaneous=True)

    return y


def get_individual_prediction_expression(model: Model):
    """Get the full symbolic expression for the modelled individual prediction

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_individual_prediction_expression
    >>> model = load_example_model("pheno_linear")
    >>> get_individual_prediction_expression(model)
    D_ETA1*(ETA_1 - OETA1) + D_ETA2*(ETA_2 - OETA2) + OPRED

    See Also
    --------
    get_population_prediction_expression : Get full symbolic epression for the population prediction
    """
    return subs(
        get_observation_expression(model),
        {sympy.Symbol(eps): 0 for eps in model.random_variables.epsilons.names},
        simultaneous=True,
    )


def get_population_prediction_expression(model: Model):
    """Get the full symbolic expression for the modelled population prediction

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_population_prediction_expression
    >>> model = load_example_model("pheno_linear")
    >>> get_population_prediction_expression(model)
    -D_ETA1*OETA1 - D_ETA2*OETA2 + OPRED

    See also
    --------
    get_individual_prediction_expression : Get full symbolic epression for the individual prediction
    """

    return subs(
        get_individual_prediction_expression(model),
        {sympy.Symbol(eta): 0 for eta in model.random_variables.etas.names},
        simultaneous=True,
    )


def calculate_eta_gradient_expression(model: Model):
    """Calculate the symbolic expression for the eta gradient

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, calculate_eta_gradient_expression
    >>> model = load_example_model("pheno_linear")
    >>> calculate_eta_gradient_expression(model)
    [D_ETA1, D_ETA2]

    See also
    --------
    calculate_epsilon_gradient_expression : Epsilon gradient
    """
    y = get_individual_prediction_expression(model)
    d = [y.diff(sympy.Symbol(x)) for x in model.random_variables.etas.names]
    return d


def calculate_epsilon_gradient_expression(model: Model):
    """Calculate the symbolic expression for the epsilon gradient

    This function currently only support models without ODE systems

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Expression
        Symbolic expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, calculate_epsilon_gradient_expression
    >>> model = load_example_model("pheno_linear")
    >>> calculate_epsilon_gradient_expression(model)
    [D_EPS1 + D_EPSETA1_1*(ETA_1 - OETA1) + D_EPSETA1_2*(ETA_2 - OETA2)]

    See also
    --------
    calculate_eta_gradient_expression : Eta gradient
    """

    y = get_observation_expression(model)
    d = [y.diff(sympy.Symbol(x)) for x in model.random_variables.epsilons.names]
    return d


def create_symbol(model: Model, stem: str, force_numbering: bool = False):
    """Create a new unique variable symbol given a model

    Parameters
    ----------
    model : Model
        Pharmpy model object
    stem : str
        First part of the new variable name
    force_numbering : bool
        Forces addition of number to name even if variable does not exist, e.g.
        COVEFF → COVEFF1

    Returns
    -------
    Symbol
        Created symbol with unique name

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_symbol
    >>> model = load_example_model("pheno")
    >>> create_symbol(model, "TEMP")
    TEMP
    >>> create_symbol(model, "TEMP", force_numbering=True)
    TEMP1
    >>> create_symbol(model, "CL")
    CL1
    """
    return _create_symbol(
        model.statements,
        model.parameters,
        model.random_variables,
        model.datainfo,
        stem,
        force_numbering,
    )


def _create_symbol(statements, parameters, random_variables, datainfo, stem, force_numbering):
    symbols = [str(symbol) for symbol in statements.free_symbols]
    params = [param.name for param in parameters]
    rvs = random_variables.names
    dataset_col = datainfo.names

    all_names = symbols + params + rvs + dataset_col

    if str(stem) not in all_names and not force_numbering:
        return sympy.Symbol(str(stem))

    i = 1
    while True:
        candidate = f'{stem}{i}'
        if candidate not in all_names:
            return sympy.Symbol(candidate)
        i += 1


def _find_eta_assignments(model):
    # NOTE: This locates all assignment to ETAs of symbols that do not depend on
    # any other ETA
    statements = model.statements.before_odes
    etas = {sympy.Symbol(eta) for eta in model.random_variables.etas.names}
    found = set()
    leafs = []

    for i, s in reversed(list(enumerate(statements))):
        if (
            s.symbol not in found
            and not etas.isdisjoint(s.free_symbols)
            and len(etas & statements[:i].full_expression(s.expression).free_symbols) == 1
        ):
            leafs.append((i, s))
            found.update(s.free_symbols)

    return reversed(leafs)


def mu_reference_model(model: Model):
    r"""Convert model to use mu-referencing

    Mu-referencing an eta is to separately define its actual mu (mean) parameter.
    For example: :math:`CL = \theta_1 e^{\eta_1}` with :math:`\eta_1` following a zero-mean
    normal distribution would give :math:`\mu_1 = \log{\theta_1}` and
    :math:`CL = e^{\mu_1 + \eta_1}`

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Pharmpy model object

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, mu_reference_model
    >>> model = load_example_model("pheno")
    >>> model = mu_reference_model(model)
    >>> model.statements.before_odes
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
    TVV = PTVV⋅WGT
          ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩       TVV           otherwise
    μ₁ = log(TVCL)
          ETA₁ + μ₁
    CL = ℯ
    μ₂ = log(TVV)
         ETA₂ + μ₂
    V = ℯ
    S₁ = V
    """
    index = {sympy.Symbol(eta): i for i, eta in enumerate(model.random_variables.etas.names, 1)}
    etas = set(index)

    offset = 0

    statements = model.statements
    for old_ind, assignment in _find_eta_assignments(model):
        # NOTE: The sequence of old_ind must be increasing
        eta = next(iter(etas.intersection(assignment.expression.free_symbols)))
        old_def = assignment.expression
        dep = old_def.as_independent(eta)[1]
        mu = sympy.Symbol(f'mu_{index[eta]}')
        if mu in old_def.free_symbols:
            # If mu reference is already used, ignore
            pass
        else:
            new_def = subs(dep, {eta: mu + eta})
            mu_expr = sympy.solve(old_def - new_def, mu)[0]
            insertion_ind = offset + old_ind
            statements = (
                statements[0:insertion_ind]
                + Assignment(mu, mu_expr)
                + Assignment(assignment.symbol, new_def)
                + statements[insertion_ind + 1 :]
            )
            offset += 1  # NOTE: We need this offset because we replace one
            # statement by two statements
    model = model.replace(statements=statements).update_source()
    return model


def simplify_expression(model: Model, expr: Union[int, float, str, sympy.Expr]):
    """Simplify expression given constraints in model

    Parameters
    ----------
    model : Model
        Pharmpy model object
    expr : Expression
        Expression to simplify

    Returns
    -------
    Expression
        Simplified expression

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, simplify_expression
    >>> model = load_example_model("pheno")
    >>> simplify_expression(model, "Abs(PTVCL)")
    PTVCL
    """
    return _simplify_expression_from_parameters(expr, model.parameters)


def _simplify_expression_from_parameters(expr, parameters):
    expr = parse_expr(expr)
    d = {}
    for p in parameters:
        if p.fix:
            s = sympy.Float(p.init)
        elif p.upper < 0:
            s = sympy.Symbol(p.name, real=True, negative=True)
            d[s] = p.symbol
        elif p.upper <= 0:
            s = sympy.Symbol(p.name, real=True, nonpositive=True)
            d[s] = p.symbol
        elif p.lower > 0:
            s = sympy.Symbol(p.name, real=True, positive=True)
            d[s] = p.symbol
        elif p.lower >= 0:
            s = sympy.Symbol(p.name, real=True, nonnegative=True)
            d[s] = p.symbol
        else:
            s = sympy.Symbol(p.name, real=True)
            d[s] = p.symbol
        expr = subs(expr, {p.symbol: s})
    # Remaining symbols should all be real
    for s in expr.free_symbols:
        if s.is_real is not True:
            new = sympy.Symbol(s.name, real=True)
            expr = subs(expr, {s: new})
            d[new] = s
    simp = subs(sympy.simplify(expr), d)  # Subs symbols back to non-constrained
    return simp


def make_declarative(model: Model):
    """Make the model statments declarative

    Each symbol will only be declared once.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model.statements.before_odes
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
    TVV = PTVV⋅WGT
          ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩       TVV           otherwise
               ETA₁
    CL = TVCL⋅ℯ
             ETA₂
    V = TVV⋅ℯ
    S₁ = V
    >>> model = make_declarative(model)
    >>> model.statements.before_odes
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
          ⎧PTVV⋅WGT⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩      PTVV⋅WGT          otherwise
               ETA₁
    CL = TVCL⋅ℯ
             ETA₂
    V = TVV⋅ℯ
    S₁ = V
    """
    assigned_symbols = set()
    duplicated_symbols = {}  # symbol to last index
    for i, s in enumerate(model.statements):
        if not isinstance(s, Assignment):
            continue
        symb = s.symbol
        if symb in assigned_symbols:
            if symb not in duplicated_symbols:
                duplicated_symbols[symb] = []
            duplicated_symbols[symb].append(i)
        else:
            assigned_symbols.add(symb)

    current = {}
    newstats = []
    for i, s in enumerate(model.statements):
        if not isinstance(s, Assignment):
            s = s.subs(current)
            newstats.append(s)
        elif s.symbol in duplicated_symbols:
            if i not in duplicated_symbols[s.symbol]:
                current[s.symbol] = s.expression
            else:
                duplicated_symbols[s.symbol] = duplicated_symbols[s.symbol][1:]
                if duplicated_symbols[s.symbol]:
                    current[s.symbol] = subs(s.expression, current)
                else:
                    ass = Assignment(s.symbol, subs(s.expression, current))
                    newstats.append(ass)
                    del current[s.symbol]
        else:
            ass = Assignment(s.symbol, subs(s.expression, current))
            newstats.append(ass)

    model = model.replace(statements=Statements(newstats))
    return model.update_source()


def cleanup_model(model: Model):
    """Perform various cleanups of a model

    This is what is currently done

    * Make model statements declarative, i.e. only one assignment per symbol
    * Inline all assignments of one symbol, e.g. X = Y

    Notes
    -----
    When creating NONMEM code from the cleaned model Pharmpy might need to
    add certain assignments to make it in line with what NONMEM requires.

    Parameters
    ----------
    model : Model
        Pharmpy model object

    Returns
    -------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model.statements
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
    TVV = PTVV⋅WGT
          ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩       TVV           otherwise
               ETA₁
    CL = TVCL⋅ℯ
             ETA₂
    V = TVV⋅ℯ
    S₁ = V
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘
        A_CENTRAL(t)
        ────────────
    F =      S₁
    W = F
    Y = EPS₁⋅W + F
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W
    >>> model = cleanup_model(model)
    >>> model.statements
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
          ⎧PTVV⋅WGT⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩      PTVV⋅WGT           otherwise
               ETA₁
    CL = TVCL⋅ℯ
             ETA₂
    V = TVV⋅ℯ
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘
        A_CENTRAL(t)
        ────────────
    F =      V
    Y = EPS₁⋅F + F
    IRES = DV - F
            IRES
            ────
    IWRES =  F
    """
    model = make_declarative(model)

    current = {}
    newstats = []
    for s in model.statements:
        if isinstance(s, Assignment) and s.expression.is_Symbol:
            current[s.symbol] = s.expression
        else:
            n = s.subs(current)
            newstats.append(n)

    model = model.replace(statements=Statements(newstats))
    return model.update_source()


def greekify_model(model: Model, named_subscripts: bool = False):
    """Convert to using greek letters for all population parameters

    Parameters
    ----------
    model : Model
        Pharmpy model
    named_subscripts : bool
        Use previous parameter names as subscripts. Default is to use integer subscripts

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model.statements
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = PTVCL⋅WGT
    TVV = PTVV⋅WGT
          ⎧TVV⋅(THETA₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩      TVV           otherwise
               ETA₁
    CL = TVCL⋅ℯ
             ETA₂
    V = TVV⋅ℯ
    S₁ = V
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘
        A_CENTRAL(t)
        ────────────
    F =      S₁
    W = F
    Y = EPS₁⋅W + F
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W

    >>> model = greekify_model(cleanup_model(model))
    >>> model.statements
            ⎧TIME  for AMT > 0
            ⎨
    BTIME = ⎩ 0     otherwise
    TAD = -BTIME + TIME
    TVCL = WGT⋅θ₁
          ⎧WGT⋅θ₂⋅(θ₃ + 1)  for APGR < 5
          ⎨
    TVV = ⎩    WGT⋅θ₂        otherwise
               η₁
    CL = TVCL⋅ℯ
             η₂
    V = TVV⋅ℯ
    Bolus(AMT, admid=1) → CENTRAL
    ┌───────┐
    │CENTRAL│──CL/V→
    └───────┘
        A_CENTRAL(t)
        ────────────
    F =      V
    Y = F⋅ε₁ + F
    IRES = DV - F
            IRES
            ────
    IWRES =  F

    """

    def get_subscript(param, i, named_subscripts):
        if named_subscripts:
            if isinstance(param, str):
                subscript = param
            else:
                subscript = param.name
        else:
            subscript = i
        return subscript

    def get_2d_subscript(param, row, col, named_subscripts):
        if named_subscripts:
            subscript = param.name
        else:
            subscript = f'{row}{col}'
        return subscript

    subs = {}
    for i, theta in enumerate(get_thetas(model), start=1):
        subscript = get_subscript(theta, i, named_subscripts)
        subs[theta.symbol] = sympy.Symbol(f"theta_{subscript}")
    omega = model.random_variables.covariance_matrix
    for row in range(omega.rows):
        for col in range(omega.cols):
            if col > row:
                break
            elt = omega[row, col]
            if elt == 0:
                continue
            subscript = get_2d_subscript(elt, row + 1, col + 1, named_subscripts)
            subs[elt] = sympy.Symbol(f"omega_{subscript}")
    sigma = model.random_variables.covariance_matrix
    for row in range(sigma.rows):
        for col in range(sigma.cols):
            if col > row:
                break
            elt = sigma[row, col]
            if elt == 0:
                continue
            subscript = get_2d_subscript(elt, row + 1, col + 1, named_subscripts)
            subs[elt] = sympy.Symbol(f"sigma_{subscript}")
    for i, eta in enumerate(model.random_variables.etas.names, start=1):
        subscript = get_subscript(eta, i, named_subscripts)
        subs[sympy.Symbol(eta)] = sympy.Symbol(f"eta_{subscript}")
    for i, epsilon in enumerate(model.random_variables.epsilons.names, start=1):
        subscript = get_subscript(epsilon, i, named_subscripts)
        subs[sympy.Symbol(epsilon)] = sympy.Symbol(f"epsilon_{subscript}")
    from pharmpy.modeling import rename_symbols

    model = rename_symbols(model, subs)
    return model


def get_individual_parameters(
    model: Model,
    level: Literal['iiv', 'iov', 'random', 'all'] = 'all',
    dv: Union[sympy.Symbol, str, int, None] = None,
) -> List[str]:
    """Retrieves all individual parameters in a :class:`pharmpy.model`.

    By default all individual parameters will be found even ones having no random effect. The level
    arguments makes it possible to find only those having any random effect or only those having a certain
    random effect. Using the dv option will give all individual parameters affecting a certain dv. Note that
    the DV for PD in a PKPD model often also is affected by the PK parameters.

    Parameters
    ----------
    model : Model
        Pharmpy model to retrieve the individuals parameters from
    level : {'iiv', 'iov', 'random', 'all'}
        The variability level to look for: 'iiv', 'iov', 'random' or 'all' (default)
    dv : Union[sympy.Symbol, str, int, None]
        Name or DVID of dependent variable. None for all (default)

    Return
    ------
    list[str]
        A list of the parameter names as strings

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_individual_parameters(model)
    ['CL', 'V']
    >>> get_individual_parameters(model, 'iiv')
    ['CL', 'V']
    >>> get_individual_parameters(model, 'iov')
    []

    See also
    --------
    get_pd_parameters
    get_pk_parameters
    get_rv_parameters
    has_random_effect

    """
    # FIXME: Support multiple DVs

    model = make_declarative(model)
    model = _replace_trivial_redefinitions(model)
    statements = model.statements

    # Arrow means is depending on
    full_graph = statements._create_dependency_graph()
    all_statements = set(full_graph.nodes)
    theta_eta_deps = _find_theta_eta_dependents(model, full_graph)
    eps_deps = _find_eps_dependents(model, full_graph)
    separable_statements = _find_separable_statements(model)
    ode_statement = _find_ode_statement(model)
    dv_statements = _find_dv_statements(model)
    time_deps = _find_time_dependents(model, full_graph)
    all_candidates = (
        all_statements.intersection(theta_eta_deps)
        - eps_deps
        - separable_statements
        - ode_statement
        - dv_statements
        - time_deps
    )

    parameter_symbs = set()
    if dv is None:
        dvs = model.dependent_variables.keys()
    else:
        dvs = {get_dv_symbol(model, dv)}
    for y in dvs:
        ind = statements.find_assignment_index(y)
        gsub = _subgraph_of(full_graph, ind)
        gsub = _cut_partial_odes(model, gsub, ind)

        candidates = set(gsub.nodes).intersection(all_candidates)

        parameter_indices = _find_individual_parameters(gsub, candidates)

        parameter_symbs |= {statements[i].symbol for i in parameter_indices}

    if level == 'random':
        wanted_levels = {'iov', 'iiv'}
    elif level == 'iiv':
        wanted_levels = {'iiv'}
    elif level == 'iov':
        wanted_levels = {'iov'}
    else:
        wanted_levels = None

    if wanted_levels is not None:
        filtered = set()
        for param in parameter_symbs:
            levels = _random_levels_of_parameter(model, full_graph, param)
            if not wanted_levels.isdisjoint(levels):
                filtered.add(param)
    else:
        filtered = parameter_symbs

    return sorted([s.name for s in filtered])


def _subgraph_of(g, ind):
    gsub = g.subgraph(nx.descendants(g, ind) | {ind})
    return gsub


def _cut_partial_odes(model, g, dv):
    # Cut not needed deps of the ode-system for this dv
    # This prevents the entire ode-system to be needed in case
    # of a disjoint system.

    odes = model.statements.ode_system

    if odes is None:
        return g

    def dep_funcs(eq):
        from sympy.core.function import AppliedUndef

        funcs = eq.atoms(AppliedUndef)
        return funcs

    def dep_amounts(ode_deps, amounts):
        deps = set(amounts)
        for amount in amounts:
            deps |= set(nx.descendants(ode_deps, amount))
        return deps

    ode_deps = nx.DiGraph()
    for eq in odes.eqs:
        fn = dep_funcs(eq.lhs).pop()
        ode_deps.add_node(fn)

    for eq in odes.eqs:
        lhs_func = dep_funcs(eq.lhs).pop()
        rhs_funcs = dep_funcs(eq.rhs) - {lhs_func}
        for fn in rhs_funcs:
            ode_deps.add_edge(lhs_func, fn)

    all_dep_funcs = set()
    for i in g.nodes():
        s = model.statements[i]
        if isinstance(s, Assignment):
            all_dep_funcs |= dep_funcs(s.expression)

    deps = dep_amounts(ode_deps, all_dep_funcs)

    dep_symbs = set()
    nondep_symbs = set()

    for eq in odes.eqs:
        lhs_func = dep_funcs(eq.lhs).pop()
        if lhs_func in deps:
            dep_symbs |= eq.free_symbols
        else:
            nondep_symbs |= eq.free_symbols

    not_needed_symbs = nondep_symbs - dep_symbs

    for i, s in enumerate(model.statements):
        if isinstance(s, CompartmentalSystem):
            ode_index = i
            break

    g = g.copy()
    for i, s in enumerate(model.statements):
        if i in g and isinstance(s, Assignment) and s.symbol in not_needed_symbs:
            try:
                g.remove_edge(ode_index, i)
            except nx.NetworkXError:
                pass
            try:
                g.remove_edge(i, ode_index)
            except nx.NetworkXError:
                pass
    g = _subgraph_of(g, dv)  # Remove all nodes no longer connected
    return g


def _random_levels_of_parameter(model, g, param):
    ind = model.statements.find_assignment_index(param)
    iiv_symbs = model.random_variables.iiv.symbols
    iov_symbs = model.random_variables.iov.symbols
    levels = set()
    for node in nx.dfs_preorder_nodes(g, ind):
        if not model.statements[node].rhs_symbols.isdisjoint(iiv_symbs):
            levels.add('iiv')
        if not model.statements[node].rhs_symbols.isdisjoint(iov_symbs):
            levels.add('iov')
    return sorted(levels)


def _find_ode_statement(model):
    for i, s in enumerate(model.statements):
        if isinstance(s, CompartmentalSystem):
            return {i}
    return set()


def _find_dv_statements(model):
    inds = set()
    for i, s in enumerate(model.statements):
        if isinstance(s, Assignment) and s.symbol in model.dependent_variables.keys():
            inds.add(i)
    return inds


def _find_separable_statements(model):
    # Statements with only assignment dependencies
    # where these are dependencies of other assignments
    statements = model.statements
    all_assigned_symbols = _get_all_assigned_symbols(model)
    candidates = {
        i
        for i, s in enumerate(statements)
        if isinstance(s, Assignment) and s.rhs_symbols.issubset(all_assigned_symbols)
    }
    separable = set()
    for i in candidates:
        rhs = statements[i].rhs_symbols
        for j, s in enumerate(statements):
            if i != j and not rhs.isdisjoint(s.rhs_symbols):
                separable.add(i)
                break
    return separable


def _get_all_assigned_symbols(model):
    statements = model.statements
    all_assigned_symbols = {
        statement.symbol for statement in statements if isinstance(statement, Assignment)
    }
    return all_assigned_symbols


def _replace_trivial_redefinitions(model):
    # Remove and replace X = Y and X = 1/Y
    statements = model.statements
    all_assigned_symbols = _get_all_assigned_symbols(model)

    d = {}
    keep = []
    for s in statements:
        if isinstance(s, Assignment) and (
            s.expression in all_assigned_symbols or 1 / s.expression in all_assigned_symbols
        ):
            d[s.symbol] = s.expression
        else:
            keep.append(s)
    new = Statements(tuple(keep)).subs(d)
    return model.replace(statements=new)


def _find_individual_parameters(g, candidates):
    # Starting from all sinks of g find all deepest candidate nodes
    keep = set()
    current_step = _sinks(g)

    while current_step:
        next_step = set()
        for node in current_step:
            pred = set(g.predecessors(node))
            if node in candidates:
                if pred.isdisjoint(candidates):
                    keep.add(node)
                else:
                    next_step.update(pred)
            else:
                next_step.update(pred)
        current_step = next_step
    return keep


def _sinks(g):
    return (node for node, out_degree in g.out_degree() if out_degree == 0)


def _find_time_dependents(model, g):
    direct = _find_direct_time_dependents(model)
    return _find_dependents(g, direct)


def _find_dependents(g, nodes):
    deps = nodes.copy()
    for i in nodes:
        ancestors = nx.ancestors(g, i)
        deps.update(ancestors)
    return deps


def _find_direct_param_dependents(model, g, eps=False):
    # Look for statements that are directly depending on
    # either (thetas, omegas or etas) or (sigmas or epsilons)
    deps = set()
    if not eps:
        thetas_and_omegas = get_thetas(model) + get_omegas(model)
        pop_params = set(thetas_and_omegas.symbols)
        rvs = set(model.random_variables.etas.symbols)
    else:
        sigmas = get_sigmas(model)
        pop_params = set(sigmas.symbols)
        rvs = set(model.random_variables.epsilons.symbols)
    all_symbs = pop_params | rvs
    for i in g.nodes:
        s = model.statements[i]
        if not all_symbs.isdisjoint(s.rhs_symbols):
            deps.add(i)
    return deps


def _find_theta_eta_dependents(model, g):
    direct = _find_direct_param_dependents(model, g)
    return _find_dependents(g, direct)


def _find_eps_dependents(model, g):
    direct = _find_direct_param_dependents(model, g, eps=True)
    return _find_dependents(g, direct)


def _find_direct_time_dependents(model):
    statements = model.statements
    odes = statements.ode_system
    if odes is None:
        return set()
    return {i for i, s in enumerate(statements) if odes.t in s.rhs_symbols}


def depends_on(model: Model, symbol: str, other: str):
    return _depends_on_any_of(
        model.statements.before_odes, sympy.Symbol(symbol), [sympy.Symbol(other)]
    )


def _depends_on_any_of(
    assignments: Statements, symbol: sympy.Symbol, symbols: Iterable[sympy.Symbol]
):
    dependency_graph = _dependency_graph(assignments)
    if symbol not in dependency_graph:
        raise KeyError(symbol)

    # NOTE: Could be faster by returning immediately once found
    return not reachable_from({symbol}, lambda x: dependency_graph.get(x, [])).isdisjoint(symbols)


def has_random_effect(
    model: Model, parameter: str, level: Literal['iiv', 'iov', 'all'] = 'all'
) -> bool:
    """Decides whether the given parameter of a :class:`pharmpy.model` has a
    random effect.

    Parameters
    ----------
    model : Model
        Input Pharmpy model
    parameter: str
        Input parameter
    level : {'iiv', 'iov', 'all'}
        The variability level to look for: 'iiv', 'iov', or 'all' (default)

    Return
    ------
    bool
        Whether the given parameter has a random effect

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> has_random_effect(model, 'S1')
    True
    >>> has_random_effect(model, 'CL', 'iiv')
    True
    >>> has_random_effect(model, 'CL', 'iov')
    False

    See also
    --------
    get_individual_parameters
    get_rv_parameters

    """

    rvs = _rvs(model, level)
    symbol = sympy.Symbol(parameter)
    return _depends_on_any_of(model.statements.before_odes, symbol, map(sympy.Symbol, rvs.names))


def get_rv_parameters(model: Model, rv: str) -> List[str]:
    """Retrieves parameters in :class:`pharmpy.model.Model` given a random variable.

    Parameters
    ----------
    model : Model
        Pharmpy model to retrieve parameters from
    rv : str
        Name of random variable to retrieve

    Return
    ------
    list[str]
        A list of parameter names for the given random variable

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_rv_parameters(model, 'ETA_1')
    ['CL']

    See also
    --------
    has_random_effect
    get_pk_parameters
    get_individual_parameters

    """
    if rv not in model.random_variables.names:
        raise ValueError(f'Could not find random variable: {rv}')

    natural_assignments = _get_natural_assignments(model.statements.before_odes)

    free_symbols = model.statements.free_symbols
    dependency_graph = _dependency_graph(natural_assignments)
    return sorted(map(str, _filter_symbols(dependency_graph, free_symbols, {sympy.Symbol(rv)})))


def get_parameter_rv(
    model: Model, parameter: str, var_type: Literal['iiv', 'iov'] = 'iiv'
) -> List[str]:
    """Retrieves name of random variable in :class:`pharmpy.model.Model` given a parameter.

    Parameters
    ----------
    model : Model
        Pharmpy model to retrieve parameters from
    parameter : str
        Name of parameter to retrieve random variable from
    var_type: {'iiv', 'iov'}
        Variability type: iiv (default) or iov

    Return
    ------
    list[str]
        A list of random variable names for the given parameter

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_parameter_rv(model, 'CL')
    ['ETA_1']

    See also
    --------
    get_rv_parameters
    has_random_effect
    get_pk_parameters
    get_individual_parameters

    """
    if parameter not in list(map(str, model.statements.free_symbols)):
        raise ValueError(f'Could not find parameter {parameter}')
    if parameter in model.random_variables.names:
        raise ValueError(f"{parameter} is a random variable. Only parameters are accepted as input")

    natural_assignments = _get_natural_assignments(model.statements.before_odes)

    rv = list(
        map(
            lambda rv_string: sympy.Symbol(rv_string),
            getattr(model.random_variables, var_type).names,
        )
    )

    dependency_graph = graph_inverse(_dependency_graph(natural_assignments))
    return sorted(map(str, _filter_symbols(dependency_graph, rv, {sympy.Symbol(parameter)})))


@dataclass(frozen=True)
class AssignmentGraphNode:
    expression: sympy.Expr
    index: int
    previous: Dict[sympy.Symbol, AssignmentGraphNode]


def _make_assignments_graph(statements: Statements) -> Dict[sympy.Symbol, AssignmentGraphNode]:
    last_assignments: Dict[sympy.Symbol, AssignmentGraphNode] = {}

    for i, statement in enumerate(statements):
        if not isinstance(statement, Assignment):
            continue

        node = AssignmentGraphNode(
            statement.expression,
            i,
            {
                symbol: last_assignments[symbol]
                for symbol in statement.expression.free_symbols
                if symbol in last_assignments
            },
        )

        last_assignments[statement.symbol] = node

    return last_assignments


def remove_covariate_effect_from_statements(
    model: Model, before_odes: Statements, parameter: str, covariate: str
) -> Iterable[Statement]:
    assignments = _make_assignments_graph(before_odes)

    thetas = _theta_symbols(model)

    new_before_odes = list(before_odes)

    symbol = sympy.Symbol(parameter)
    graph_node = assignments[symbol]

    tree_node = _remove_covariate_effect_from_statements_recursive(
        thetas,
        graph_node.previous,
        new_before_odes,
        symbol,
        graph_node.expression,
        sympy.Symbol(covariate),
        None,
    )

    assert tree_node.changed
    assert tree_node.contains_theta

    if tree_node.changed:
        new_before_odes[graph_node.index] = Assignment(
            sympy.Symbol(parameter), tree_node.expression
        )

    return new_before_odes


def _neutral(expr: sympy.Expr) -> sympy.Integer:
    if isinstance(expr, sympy.Add):
        return sympy.Integer(0)
    if isinstance(expr, sympy.Mul):
        return sympy.Integer(1)
    if isinstance(expr, sympy.Pow):
        return sympy.Integer(1)

    raise ValueError(f'{type(expr)}: {repr(expr)} ({expr.free_symbols})')


def _theta_symbols(model: Model) -> Set[sympy.Symbol]:
    rvs_fs = model.random_variables.free_symbols
    return {p.symbol for p in model.parameters if p.symbol not in rvs_fs}


def _depends_on_any(symbols: Set[sympy.Symbol], expr: sympy.Expr) -> bool:
    return any(map(lambda s: s in symbols, expr.free_symbols))


def _is_constant(thetas: Set[sympy.Symbol], expr: sympy.Expr) -> bool:
    return all(map(lambda s: s in thetas, expr.free_symbols))


def _is_univariate(thetas: Set[sympy.Symbol], expr: sympy.Expr, variable: sympy.Symbol) -> bool:
    return all(map(lambda s: s in thetas, expr.free_symbols - {variable}))


def simplify_model(
    model: Model, old_statements: Iterable[Statement], statements: Iterable[Statement]
):
    odes = model.statements.ode_system
    fs = odes.free_symbols if odes is not None else set()
    old_fs = fs.copy()

    kept_statements_reversed = []

    for old_statement, statement in reversed(list(zip(old_statements, statements))):
        if not isinstance(statement, Assignment):
            kept_statements_reversed.append(statement)
            continue

        assert isinstance(old_statement, Assignment)

        if (old_statement == statement and statement.symbol not in old_fs) or (
            statement.symbol in fs and statement.symbol != statement.expression
        ):
            kept_statements_reversed.append(statement)
            fs.discard(statement.symbol)
            fs.update(statement.expression.free_symbols)

        old_fs.discard(old_statement.symbol)
        old_fs.update(old_statement.expression.free_symbols)

    kept_thetas = fs.intersection(_theta_symbols(model))
    kept_statements = list(reversed(kept_statements_reversed))

    return kept_thetas, kept_statements


@dataclass(frozen=True)
class ExpressionTreeNode:
    expression: sympy.Expr
    changed: bool
    constant: bool
    contains_theta: bool


def _full_expression(assignments: Dict[sympy.Symbol, AssignmentGraphNode], expr: sympy.Expr):
    return expr.xreplace(
        {
            symbol: _full_expression(node.previous, node.expression)
            for symbol, node in assignments.items()
        }
    )


def _remove_covariate_effect_from_statements_recursive(
    thetas: Set[sympy.Symbol],
    assignments: Dict[sympy.Symbol, AssignmentGraphNode],
    statements: List[Assignment],
    symbol: sympy.Symbol,
    expression: sympy.Expr,
    covariate: sympy.Symbol,
    parent: Union[None, sympy.Expr],
) -> ExpressionTreeNode:
    if not expression.args:
        if expression in assignments:
            # NOTE: expression is a symbol and is defined in a previous assignment
            graph_node = assignments[expression]
            tree_node = _remove_covariate_effect_from_statements_recursive(
                thetas,
                graph_node.previous,
                statements,
                expression,
                graph_node.expression,
                covariate,
                parent,
            )
            if tree_node.changed:
                statements[graph_node.index] = Assignment(expression, tree_node.expression)

            return ExpressionTreeNode(
                expression, tree_node.changed, tree_node.constant, tree_node.contains_theta
            )

        if expression == covariate:
            # NOTE: expression is the covariate symbol for which we want to
            # remove all effects
            return ExpressionTreeNode(_neutral(parent), True, True, False)

        # NOTE: Other atom
        return ExpressionTreeNode(
            expression, False, _is_constant(thetas, expression), _depends_on_any(thetas, expression)
        )

    if isinstance(expression, sympy.Piecewise):
        if any(map(lambda t: covariate in t[1].free_symbols, expression.args)):
            # NOTE: At least one condition depends on the covariate
            if all(
                map(
                    lambda t: _is_univariate(
                        thetas, _full_expression(assignments, t[1]), covariate
                    ),
                    expression.args,
                )
            ):
                # NOTE: If expression is piecewise univariate and condition depends on
                # covariate, return simplest expression from cases
                expr = min(
                    (t[0] for t in expression.args),
                    key=sympy.count_ops,
                )
                tree_node = _remove_covariate_effect_from_statements_recursive(
                    thetas, assignments, statements, symbol, expr, covariate, parent
                )
                return ExpressionTreeNode(
                    tree_node.expression, True, tree_node.constant, tree_node.contains_theta
                )
            else:
                raise NotImplementedError(
                    'Cannot handle multivariate Piecewise where condition depends on covariate.'
                )

    children = list(
        map(
            lambda expr: _remove_covariate_effect_from_statements_recursive(
                thetas, assignments, statements, symbol, expr, covariate, expression
            ),
            expression.args,
        )
    )

    # TODO: Take THETA limits into account. Currently we assume any
    # offset/factor can be compensated but this is not true in general.
    can_be_scaled_or_offset = any(map(lambda n: not n.changed and n.contains_theta, children))

    changed = any(map(lambda n: n.changed, children))
    is_constant = all(map(lambda n: n.constant, children))
    contains_theta = any(map(lambda n: n.contains_theta, children))

    if not changed:
        return ExpressionTreeNode(expression, False, is_constant, contains_theta)

    if not can_be_scaled_or_offset:
        return ExpressionTreeNode(
            expression.func(*map(lambda n: n.expression, children)),
            True,
            is_constant,
            contains_theta,
        )

    return ExpressionTreeNode(
        expression.func(
            *map(
                lambda n: _neutral(expression)
                if n.changed and n.constant and n.expression != symbol
                else n.expression,
                children,
            )
        ),
        True,
        is_constant,
        contains_theta,
    )


def get_pk_parameters(
    model: Model, kind: Literal['absorption', 'distribution', 'elimination', 'all'] = 'all'
) -> List[str]:
    """Retrieves PK parameters in :class:`pharmpy.model.Model`.

    Parameters
    ----------
    model : Model
        Pharmpy model to retrieve the PK parameters from
    kind : {'absorption', 'distribution', 'elimination', 'all'}
        The type of parameter to retrieve: 'absorption', 'distribution',
        'elimination', or 'all' (default).

    Return
    ------
    list[str]
        A list of the PK parameter names of the given model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> get_pk_parameters(model)
    ['CL', 'V']
    >>> get_pk_parameters(model, 'absorption')
    []
    >>> get_pk_parameters(model, 'distribution')
    ['V']
    >>> get_pk_parameters(model, 'elimination')
    ['CL']

    See also
    --------
    get_individual_parameters
    get_rv_parameters

    """
    pkparams = get_individual_parameters(model, dv=1)
    if len(model.dependent_variables) == 2:
        dmparams = _get_drug_metabolite_parameters(model, dv=2)
        if dmparams:
            pkparams = list(set(pkparams).union(dmparams))
    model = make_declarative(model)
    model = _replace_trivial_redefinitions(model)
    if kind != 'all':
        symbs = {str(symb) for symb in _pk_free_symbols(model, kind)}
        pkparams = set(pkparams).intersection(symbs)
    return sorted(pkparams)


def _get_drug_metabolite_parameters(model, dv=2):
    statements = model.statements
    ode = statements.ode_system

    dv1_deps = [
        s
        for s in model.statements.after_odes.full_expression(get_dv_symbol(model, dv=1)).atoms(
            sympy.Function
        )
    ]
    dv1_comp = None
    dv2_deps = [
        s
        for s in model.statements.after_odes.full_expression(get_dv_symbol(model, dv=dv)).atoms(
            sympy.Function
        )
    ]
    dv2_comp = None

    comp_names = ode.compartment_names
    for comp in [ode.find_compartment(name) for name in comp_names]:
        amounts = comp.amount
        if amounts in dv1_deps:
            dv1_comp = comp
        if amounts in dv2_deps:
            dv2_comp = comp

    if dv2_comp is not None:
        if ode.get_flow(dv1_comp, dv2_comp):
            return get_individual_parameters(model, dv=2)
    else:
        None


def get_pd_parameters(model: Model) -> list[str]:
    """Retrieves PD parameters in :class:`pharmpy.model.Model`.

    Parameters
    ----------
    model : Model
        Pharmpy model to retrieve the PD parameters from

    Return
    ------
    list[str]
        A list of the PD parameter names of the given model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_direct_effect(model, "linear")
    >>> get_pd_parameters(model)
    ['B', 'SLOPE']

    See also
    --------
    get_pk_parameters

    """
    pkparams = get_individual_parameters(model, dv=1)
    pkpdparams = get_individual_parameters(model, dv=2)
    pdparams = sorted(set(pkpdparams) - set(pkparams))
    return pdparams


def _rvs(model: Model, level: str):
    if level == 'iiv':
        return model.random_variables.iiv
    if level == 'iov':
        return model.random_variables.iov
    if level == 'all':
        return model.random_variables.etas

    raise ValueError(f'Cannot handle level `{level}`')


def _get_natural_assignments(before_odes):
    # Return assignments where assignments that are constants (e.g. X=1),
    # single length expressions (e.g. S1=V), and divisions between parameters
    # (e.g. K=CL/V) have been filtered out
    classified_assignments = list(_classify_assignments(list(_assignments(before_odes))))
    natural_assignments = list(_remove_synthetic_assignments(classified_assignments))
    return natural_assignments


def _remap_compartmental_system(sset, natural_assignments):
    # Return compartmental system where rates that are synthetic assignments
    # have been substituted with their full definition (e.g K -> CL/V)
    cs = sset.ode_system

    assignments = list(_assignments(sset.before_odes))
    for assignment in reversed(assignments):
        # FIXME: Can be made more general, doesn't cover cases with recursively defined symbols (e.g. V=V/2)
        if assignment not in natural_assignments:
            # NOTE: Substitution must be made in this order
            cs = cs.subs({assignment.symbol: assignment.expression})
    return cs


def _pk_free_symbols(model, kind: str) -> Iterable[sympy.Symbol]:
    cs = model.statements.ode_system
    symbols = set()
    if kind == 'absorption':
        central = cs.central_compartment
        for dosing in cs.dosing_compartments:
            comp = dosing
            while comp != central:
                symbols |= comp.free_symbols
                comp, rate = cs.get_compartment_outflows(comp)[0]  # Assumes only one flow
                symbols |= rate.free_symbols
    elif kind == 'distribution':
        central = cs.central_compartment
        periphs = cs.find_peripheral_compartments()
        for p in periphs:
            symbols |= cs.get_flow(central, p).free_symbols
            symbols |= cs.get_flow(p, central).free_symbols
        outflow = cs.get_flow(central, output)
        cl, v = outflow.as_numer_denom()
        if v != 1:
            symbols |= {v}
        else:
            ass = model.statements.find_assignment(cl)
            cl2, v2 = ass.expression.as_numer_denom()
            if v2 != 1:
                symbols |= {v2}
    elif kind == 'elimination':
        central = cs.central_compartment
        outflow = cs.get_flow(central, output)
        cl, v = outflow.as_numer_denom()
        if v != 1:
            symbols |= {cl}
        else:
            ass = model.statements.find_assignment(cl)
            cl2, v2 = ass.expression.as_numer_denom()
            if v2 != 1:
                symbols |= {cl2}
            else:
                symbols |= {cl}
    else:
        raise ValueError(f'Cannot handle kind `{kind}`')
    return symbols


def _pk_free_symbols_from_compartment(
    cs: CompartmentalSystem, compartment: Compartment
) -> Iterable[sympy.Symbol]:
    vertices = _get_component(cs, compartment)
    edges = _get_component_edges(cs, vertices)
    is_central = compartment == cs.central_compartment
    return _get_component_free_symbols(is_central, vertices, edges)


def _get_component(cs: CompartmentalSystem, compartment: Compartment) -> Set[Compartment]:
    central_component_vertices = strongly_connected_component_of(
        cs.central_compartment,
        lambda u: map(lambda flow: flow[0], cs.get_compartment_outflows(u)),
        lambda u: map(lambda flow: flow[0], cs.get_compartment_inflows(u)),
    )

    if compartment == cs.central_compartment:
        return central_component_vertices

    flows = cs.get_compartment_inflows if compartment == output else cs.get_compartment_outflows

    return reachable_from(
        {compartment},
        lambda u: filterfalse(
            central_component_vertices.__contains__,
            map(lambda flow: flow[0], flows(u)),
        ),
    )


def _get_component_edges(cs: CompartmentalSystem, vertices: Set[Compartment]):
    return (
        ((u, v, rate) for v in vertices for u, rate in cs.get_compartment_inflows(v))
        if output in vertices
        else ((u, v, rate) for u in vertices for v, rate in cs.get_compartment_outflows(u))
    )


def _get_component_free_symbols(
    is_central: bool,
    vertices: Set[Compartment],
    edges: Iterable[Tuple[Compartment, Compartment, sympy.Expr]],
) -> Iterable[sympy.Symbol]:
    for u, v, rate in edges:
        # NOTE: These must not necessarily be outgoing edges
        assert u in vertices or v in vertices

        if u not in vertices or v not in vertices:
            # NOTE: This handles splitting the rate K = CL / V
            if len(rate.free_symbols) == 2:
                a, b = rate.free_symbols
                if rate == a / b:
                    yield a if v in vertices else b
                    continue
                elif rate == b / a:
                    yield b if v in vertices else a
                    continue

        if (u in vertices and v in vertices) or not is_central:
            # NOTE: This handles all internal edges, and in/out rates (KA, CL/V)
            yield from rate.free_symbols

    for node in vertices:
        if node == output:
            yield from set()
        else:
            yield from node.free_symbols


def _assignments(sset: Statements):
    return filter(lambda statement: isinstance(statement, Assignment), sset)


def _filter_symbols(
    dependency_graph: Dict[sympy.Symbol, Set[sympy.Symbol]],
    roots: Set[sympy.Symbol],
    leaves: Union[Set[sympy.Symbol], None] = None,
) -> Set[sympy.Symbol]:
    dependents = graph_inverse(dependency_graph)

    free_symbols = reachable_from(roots, lambda x: dependency_graph.get(x, []))

    reachable = (
        free_symbols
        if leaves is None
        else (
            reachable_from(
                leaves,
                lambda x: dependents.get(x, []),
            ).intersection(free_symbols)
        )
    )

    return reachable.difference(dependents.keys()).intersection(dependency_graph.keys())


def _classify_assignments(assignments: Sequence[Assignment]):
    dependencies = _dependency_graph(assignments)

    # Keep all symbols that have dependencies (e.g. remove constants X=1)
    symbols = set(filter(dependencies.__getitem__, dependencies.keys()))

    for assignment in assignments:
        symbol = assignment.symbol
        expression = assignment.expression
        fs = expression.free_symbols

        if symbol not in fs:  # NOTE: We skip redefinitions (e.g. CL=CL+1)
            if sympy.Symbol('t') in fs:  # FIXME: Should use ode.t here at some point
                yield 'synthetic', assignment
                continue
            elif len(fs) == 1:
                a = next(iter(fs))
                if a in symbols:
                    yield 'synthetic', assignment  # E.g. S1=V
                    continue
            elif len(fs) == 2:
                it = iter(fs)
                a = next(it)
                b = next(it)
                if (
                    a in symbols
                    and b in symbols
                    and (
                        expression == a / b
                        or expression == b / a
                        or expression == a * b
                        or expression == b * a
                    )
                ):
                    yield 'synthetic', assignment  # E.g. K=CL/V
                    continue

        yield 'natural', assignment


def _remove_synthetic_assignments(classified_assignments: List[Tuple[str, Assignment]]):
    assignments = []
    last_defined = {}

    for t, assignment in reversed(classified_assignments):
        if t == 'synthetic':
            substitution_starts_at_index = last_defined.get(assignment.symbol, 0)
            assignments = [
                succeeding
                if i < substitution_starts_at_index
                else Assignment(
                    succeeding.symbol,
                    subs(
                        succeeding.expression,
                        {assignment.symbol: assignment.expression},
                        simultaneous=True,
                    ),
                )
                for i, succeeding in enumerate(assignments)
            ]
        else:
            last_defined[assignment.symbol] = len(assignments)
            assignments.append(assignment)

    return reversed(assignments)


def _dependency_graph(assignments: Sequence[Assignment]):
    dependencies = {}

    for assignment in assignments:
        symbol = assignment.symbol
        fs = assignment.expression.free_symbols

        previous_def = dependencies.get(symbol)
        dependencies[symbol] = fs

        if previous_def is not None:
            # NOTE: This handles redefinition of symbols by expanding
            # the previous definition of symbol into existing definitions
            for key, value in dependencies.items():
                if symbol in value:
                    dependencies[key] = (value - {symbol}) | previous_def

    return dependencies


def is_real(model: Model, expr: Union[str, sympy.Expr]) -> Optional[bool]:
    """Determine if an expression is real valued given constraints of a model

    Parameters
    ----------
    model : Model
        Pharmpy model
    expr : str or expression
        Expression to test

    Return
    ------
    bool or None
        True if expression is real, False if not and None if unknown

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, is_real
    >>> import sympy
    >>> model = load_example_model("pheno")
    >>> is_real(model, "CL")
    True

    """
    expr = sympy.sympify(expr)
    return sympy.ask(sympy.Q.real(expr), assume_all(sympy.Q.real, free_images_and_symbols(expr)))


def is_linearized(model: Model):
    """Determine if a model is linearized

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    bool
        True if model has been linearized and False otherwise

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, is_linearized
    >>> model1 = load_example_model("pheno")
    >>> is_linearized(model1)
    False
    >>> model2 = load_example_model("pheno_linear")
    >>> is_linearized(model2)
    True

    """
    statements = model.statements
    if statements.ode_system is not None:
        return False
    lhs = set()
    rhs = set()
    for s in statements:
        name = s.symbol.name
        if name == 'Y':  # To support linearized frem models
            break
        lhs.add(name)
        rhs_names = {symb.name for symb in s.expression.free_symbols}
        rhs.update(rhs_names)

    for name in lhs:
        m = re.match(r'BASE|BSUM|BASE_TERMS|IPRED|ERR|ESUM|ERROR_TERMS', name)
        if not m:
            return False

    for name in rhs:
        m = re.match(
            r'D_ETA|ETA|OETA|BASE|BSUM|BASE_TERMS|OPRED|EPS|D_EPS|D_EPSETA|ERR|ESUM|ERROR_TERMS|IPRED',
            name,
        )
        if not m:
            return False

    return True


def get_dv_symbol(model: Model, dv: Union[sympy.Symbol, str, int, None] = None) -> sympy.Symbol:
    """Get the symbol for a certain dvid or dv and check that it is valid

    Parameters
    ----------
    model : Model
        Pharmpy model
    dv : Union[sympy.Symbol, str, int]
        Either a dv symbol, str or dvid. If None (default) return the
        only or first dv.

    Return
    ------
    sympy.Symbol
        DV symbol

    Example
    -------
    >>> from pharmpy.modeling import load_example_model, get_dv_symbol
    >>> model = load_example_model("pheno")
    >>> get_dv_symbol(model, "Y")
    Y
    >>> get_dv_symbol(model, 1)
    Y

    """
    if dv is None:
        dv = next(iter(model.dependent_variables))
    elif isinstance(dv, str):
        dv = sympy.Symbol(dv)
    elif isinstance(dv, int):
        for key, val in model.dependent_variables.items():
            if dv == val:
                dv = key
                break
        else:
            raise ValueError(f"DVID {dv} not defined in model")
    elif isinstance(dv, sympy.Symbol):
        pass
    else:
        raise TypeError(
            f"dv is of type {type(dv)} has to be one of Symbol or str representing "
            "a dv or int representing a dvid"
        )
    if dv not in model.dependent_variables:
        raise ValueError(f"DV {dv} not defined in model")
    return dv



from io import StringIO
from pathlib import Path

import pandas as pd
import pytest

from pharmpy.model.external.nonmem.dataset import read_nonmem_dataset
from pharmpy.modeling import (
    evaluate_epsilon_gradient,
    evaluate_eta_gradient,
    evaluate_expression,
    evaluate_individual_prediction,
    evaluate_population_prediction,
    evaluate_weighted_residuals,
)
from pharmpy.tools import read_modelfit_results

tabpath = Path(__file__).resolve().parent.parent / 'testdata' / 'nonmem' / 'pheno_real_linbase.tab'
lincorrect = read_nonmem_dataset(
    tabpath,
    ignore_character='@',
    colnames=['ID', 'G11', 'G21', 'H11', 'CIPREDI', 'DV', 'PRED', 'RES', 'WRES'],
)


def test_evaluate_expression(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_noifs.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'models' / 'pheno_noifs.mod')
    ser = evaluate_expression(model, 'TVV', res.parameter_estimates)
    assert ser[0] == pytest.approx(1.413062)
    assert ser[743] == pytest.approx(1.110262)


def test_evaluate_population_prediction(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'minimal.mod'
    model = load_model_for_test(path)

    dataset = pd.DataFrame({'ID': [1, 2], 'TIME': [0, 0], 'DV': [3, 4]})
    pred = evaluate_population_prediction(model, dataset=dataset)

    assert list(pred) == [0.1, 0.1]

    linpath = testdata / 'nonmem' / 'pheno_real_linbase.mod'
    linmod = load_model_for_test(linpath)
    pred = evaluate_population_prediction(linmod)

    pd.testing.assert_series_equal(lincorrect['PRED'], pred, rtol=1e-4, check_names=False)


def test_evaluate_individual_prediction(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'minimal.mod'
    model = load_model_for_test(path)

    dataset = read_nonmem_dataset(StringIO('1 0 3\n2 0 4\n'), colnames=['ID', 'TIME', 'DV'])
    pred = evaluate_individual_prediction(model, dataset=dataset)

    assert list(pred) == [0.1, 0.1]

    linpath = testdata / 'nonmem' / 'pheno_real_linbase.mod'
    linmod = load_model_for_test(linpath)
    res = read_modelfit_results(linpath)
    pred = evaluate_individual_prediction(model=linmod, etas=res.individual_estimates)

    pd.testing.assert_series_equal(lincorrect['CIPREDI'], pred, rtol=1e-4, check_names=False)


def test_evaluate_eta_gradient(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'minimal.mod'
    model = load_model_for_test(path)

    dataset = read_nonmem_dataset(StringIO('1 0 3\n2 0 4\n'), colnames=['ID', 'TIME', 'DV'])
    grad = evaluate_eta_gradient(model, dataset=dataset)
    pd.testing.assert_frame_equal(grad, pd.DataFrame([1.0, 1.0], columns=['dF/dETA_1']))

    linpath = testdata / 'nonmem' / 'pheno_real_linbase.mod'
    linmod = load_model_for_test(linpath)
    grad = evaluate_eta_gradient(linmod)
    pd.testing.assert_series_equal(lincorrect['G11'], grad.iloc[:, 0], rtol=1e-4, check_names=False)
    pd.testing.assert_series_equal(lincorrect['G21'], grad.iloc[:, 1], rtol=1e-4, check_names=False)


def test_evaluate_epsilon_gradient(load_model_for_test, testdata):
    path = testdata / 'nonmem' / 'minimal.mod'
    model = load_model_for_test(path)

    dataset = read_nonmem_dataset(StringIO('1 0 3\n2 0 4\n'), colnames=['ID', 'TIME', 'DV'])
    grad = evaluate_epsilon_gradient(model, dataset=dataset)
    pd.testing.assert_frame_equal(grad, pd.DataFrame([1.0, 1.0], columns=['dY/dEPS_1']))

    linpath = testdata / 'nonmem' / 'pheno_real_linbase.mod'
    linmod = load_model_for_test(linpath)
    res = read_modelfit_results(linpath)
    grad = evaluate_epsilon_gradient(linmod, etas=res.individual_estimates)
    pd.testing.assert_series_equal(lincorrect['H11'], grad.iloc[:, 0], rtol=1e-4, check_names=False)


def test_evaluate_weighted_residuals(load_model_for_test, testdata):
    linpath = testdata / 'nonmem' / 'pheno_real_linbase.mod'
    linmod = load_model_for_test(linpath)
    res = read_modelfit_results(linpath)
    wres = evaluate_weighted_residuals(linmod, parameters=dict(res.parameter_estimates))
    pd.testing.assert_series_equal(lincorrect['WRES'], wres, rtol=1e-4, check_names=False)



import re
from itertools import chain

from pharmpy.model import ModelSyntaxError

from .records.factory import create_record
from .records.raw_record import RawRecord


class Abbreviated:
    """Handling all $ABBREVIATED in a control stream"""

    def __init__(self, stream):
        self.stream = stream

    @property
    def replace(self):
        """Get all $ABBR REPLACE as a dictionary"""
        return dict(
            chain.from_iterable(
                record.replace.items() for record in self.stream.get_records('ABBREVIATED')
            )
        )

    def translate_to_pharmpy_names(self):
        return dict(
            chain.from_iterable(
                record.translate_to_pharmpy_names().items()
                for record in self.stream.get_records('ABBREVIATED')
            )
        )


class NMTranParser:
    """Parser for NMTran control streams"""

    def parse(self, text):
        records = []

        record_strings = re.split(r'^([ \t]*\$)', text, flags=re.MULTILINE)
        first = record_strings.pop(0)  # Empty if nothing before first record
        if first:
            records.append(RawRecord(first))

        for separator, s in zip(record_strings[0::2], record_strings[1::2]):
            record = create_record(separator + s)
            records.append(record)

        in_problem = False
        for record in records:
            if in_problem and record.name == 'SIZES':
                raise ModelSyntaxError('The SIZES record must come before the first PROBLEM record')
            elif record.name == 'PROBLEM':
                in_problem = True

        stream = NMTranControlStream(records=records)
        return stream


default_record_order = [
    'SIZES',
    'INPUT',
    'DATA',
    'SUBROUTINES',
    'MODEL',
    'ABBREVIATED',
    'PK',
    'PRED',
    'DES',
    'ERROR',
    'THETA',
    'OMEGA',
    'SIGMA',
    'MSFI',
    'ESTIMATION',
    'DESIGN',
    'COVARIANCE',
    'ETAS',
    'TABLE',
]


class NMTranControlStream:
    """Representation of a parsed control stream (model file)"""

    def __init__(self, records=None):
        if records is None:
            self.records = ()
        else:
            self.records = tuple(records)
        self._active_problem = 0
        self.abbreviated = Abbreviated(self)

    def get_records(self, name, problem_no=0):
        """Return a list of all records of a certain type in the current $PROBLEM"""
        current_problem = -1
        found = []
        for record in self.records:
            if record.name == 'PROBLEM':
                current_problem += 1
            if current_problem == problem_no and record.name == name:
                found.append(record)
        return found

    def _get_first_record(self, name):
        return next(iter(self.get_records(name)), None)

    def insert_record(self, record, at_index=None):
        """Create and insert a new record at the correct position

        If the record type is already present the new record will be put
        directly after the last record of that type.
        If no record of the type is present the new record will be put
        given the default record order.
        If record type is unknown. Place at the end.
        """

        name = record.name
        assert isinstance(name, str)

        if at_index:
            newrecs = self.records[0:at_index] + (record,) + self.records[at_index:]
            return NMTranControlStream(records=newrecs)

        current_problem = -1
        index = None
        for i, currec in enumerate(self.records):
            if currec.name == 'PROBLEM':
                current_problem += 1
            if current_problem == self._active_problem and currec.name == name:
                index = i

        if index is None:
            try:
                default_order_index = default_record_order.index(name)
                before_records = default_record_order[0:default_order_index]
            except ValueError:
                before_records = []
            current_problem = -1
            for i, currec in enumerate(self.records):
                if currec.name == 'PROBLEM':
                    current_problem += 1
                if current_problem == self._active_problem and currec.name in before_records:
                    index = i

        if index is None:
            index = len(self.records)

        newrecs = self.records[0 : index + 1] + (record,) + self.records[index + 1 :]
        return NMTranControlStream(records=newrecs)

    def remove_records(self, records):
        keep = [rec for rec in self.records if rec not in records]
        return NMTranControlStream(records=keep)

    def replace_records(self, old, new):
        keep = []
        first = True
        for rec in self.records:
            if rec not in old:
                keep.append(rec)
            else:
                if first:
                    keep.extend(new)
                    first = False
        return NMTranControlStream(records=keep)

    def replace_all(self, name, new):
        keep = []
        first = True
        for rec in self.records:
            if rec.name == name:
                if first:
                    keep.extend(new)
                    first = False
            else:
                keep.append(rec)
        if first:  # No record to replace. Need to insert
            index = default_record_order.index(name)
            after_index = len(keep) - 1
            for i, rec in enumerate(keep):
                try:
                    curindex = default_record_order.index(rec.name)
                except ValueError:
                    curindex = 0
                if curindex < index:
                    after_index = i
            keep = keep[0 : after_index + 1] + new + keep[after_index + 1 :]

        return NMTranControlStream(records=keep)

    def get_pred_pk_record(self):
        pred = self._get_first_record('PRED')
        if pred is not None:
            return pred

        pk = self._get_first_record('PK')
        if pk is not None:
            return pk

        raise ModelSyntaxError('Model has no $PK or $PRED')

    def get_error_pred_record(self):
        pred = self._get_first_record('PRED')
        if pred is not None:
            return pred
        error = self._get_first_record('ERROR')
        if error is not None:
            return error
        raise ModelSyntaxError('Model has no $ERROR or $PRED')

    def get_pk_record(self):
        return self._get_first_record('PK')

    def get_error_record(self):
        return self._get_first_record('ERROR')

    def get_des_record(self):
        return self._get_first_record('DES')

    def __str__(self):
        return ''.join(map(str, self.records))



from pathlib import Path

from lark import Lark, Tree, Visitor

from pharmpy.internals.parse import GenericParser, InsertMissing, with_ignored_tokens

grammar_root = Path(__file__).resolve().parent / 'grammars'


def install_grammar(cls):
    grammar = Path(grammar_root / cls.grammar_filename).resolve()
    with open(str(grammar), 'r') as fh:
        cls.lark = Lark(fh, **{**GenericParser.lark_options, **getattr(cls, 'grammar_options', {})})
    return cls


class RecordParser(GenericParser):
    pass


@install_grammar
class AbbreviatedRecordParser(RecordParser):
    grammar_filename = 'abbreviated_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (with_ignored_tokens,)


@install_grammar
class SimulationRecordParser(RecordParser):
    grammar_filename = 'simulation_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (with_ignored_tokens,)


@install_grammar
class ProblemRecordParser(RecordParser):
    grammar_filename = 'problem_record.lark'
    post_process = (
        InsertMissing(
            (
                {'root': ((0, 'raw_title'),)},
                {'raw_title': ((0, 'REST_OF_LINE'),)},
            )
        ),
    )


class InitOrLow(Visitor):
    def theta(self, tree):
        assert tree.data == 'theta'
        other = {'init', 'up'}
        subtrees = list(filter(lambda child: isinstance(child, Tree), tree.children))
        is_low = any(tree.data in other for tree in subtrees)
        for tree in subtrees:
            if tree.data == 'init_or_low':
                tree.data = 'low' if is_low else 'init'


@install_grammar
class ThetaRecordParser(RecordParser):
    grammar_filename = 'theta_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (
        InsertMissing(({'comment': ((1, 'COMMENT'),)},)),
        InitOrLow(),
        with_ignored_tokens,
    )


@install_grammar
class OmegaRecordParser(RecordParser):
    grammar_filename = 'omega_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (InsertMissing(({'comment': ((1, 'COMMENT'),)},)), with_ignored_tokens)


@install_grammar
class OptionRecordParser(RecordParser):
    grammar_filename = 'option_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (with_ignored_tokens,)


@install_grammar
class DataRecordParser(RecordParser):
    grammar_filename = 'data_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (with_ignored_tokens,)


@install_grammar
class CodeRecordParser(RecordParser):
    grammar_filename = 'code_record.lark'
    grammar_options = dict(
        propagate_positions=True,
    )
    post_process = (with_ignored_tokens,)
